/*
 * Velocity - A GNOME File Manager
 *
 *
 * Copyright (C) (2002) (Kyle Davis) <kyled@blueriver.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <glib.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <libgnomeui/gnome-icon-theme.h>
#include <libgnomeui/gnome-icon-lookup.h>
#include <libgnomevfs/gnome-vfs-mime-handlers.h>
#include "velocity-sys.h"
#include "velocity-image.h"
#include "velocity-icons.h"

#define VELOCITY_IMAGE_CACHE_SWEEP_INTERVAL_SCALE	1000	/*use seconds for our scale*/
#define VELOCITY_IMAGE_CACHE_SWEEP_INTERVAL	30	/*every 30 seconds*/
#define VELOCITY_IMAGE_CACHE_STALE_DETECT_LOW	(60/VELOCITY_IMAGE_CACHE_SWEEP_INTERVAL)	/*1 minute*/
#define VELOCITY_IMAGE_CACHE_STALE_DETECT_MEDIUM	(300/VELOCITY_IMAGE_CACHE_SWEEP_INTERVAL)	/*5 minutes*/
#define VELOCITY_IMAGE_CACHE_STALE_DETECT_HIGH	(900/VELOCITY_IMAGE_CACHE_SWEEP_INTERVAL)	/*15 minutes*/

#define VELOCITY_IMAGE_GDK_SCALE_MODE	GDK_INTERP_TILES

//TODO:  may eventually move this from global to a struct so we can have multiple cache systems
static GTree*velocity_image_cache_btree=NULL;
static guint velocity_image_cache_sweep_timer=(guint)NULL;	/*Periodic sweep of cache system*/
static GSList*velocity_image_cache_sweep_tmp;
//TODO:  keep a count on the cache size and make a configurable max cache size
//static guint velocity_image_cache_size=
static GnomeIconTheme*velocity_image_git=NULL;

static void velocity_image_cachelet_free(velocity_image_cachelet*cl){
	g_object_unref(cl->data);
	g_free(cl);
}
static gboolean velocity_image_cache_sweep_for_each_cb(const gchar*key,velocity_image_cache_object*co,gpointer user_data){
	if(co){
		guint sm[]={0,0,	/*staleness matching array (ignoring first two methods since they don't cache)*/
			VELOCITY_IMAGE_CACHE_STALE_DETECT_LOW,
			VELOCITY_IMAGE_CACHE_STALE_DETECT_MEDIUM,
			VELOCITY_IMAGE_CACHE_STALE_DETECT_HIGH};
		velocity_image_cachelet*cl;
		GSList*n=co->pixbufs;
		guint count=0;
		while(n){
			if(n->data){
				cl=(velocity_image_cachelet*)n->data;
				if(cl->staleness>=sm[(int)cl->priority]){
					GSList*t=n;
					velocity_image_cachelet_free(cl);
					cl=n->data=NULL;
					n=n->next;
					co->pixbufs=g_slist_remove(co->pixbufs,t);
					continue;
				}else{
					cl->staleness++;
					count++;
				}
			}
			n=n->next;
		}
		if(count==0){
//TODO:  small memory leak here, this is causing crashes for some reason....
//					VELOCITY_MSG("dead cachelet list(%s)[%i]",key,li);
//					if(g_slist_length(l[li])>0){
//						VELOCITY_MSG("real length is %i",g_slist_length(l[li]));
//						g_slist_free(l[li]);
//					}
			co->pixbufs=NULL;
		}
		/*if all are NULL then queue this to be dropped from our tree*/
		if(!co->pixbufs)velocity_image_cache_sweep_tmp=g_slist_prepend(velocity_image_cache_sweep_tmp,(gpointer)key);
	}
	return(FALSE);
}
static gboolean velocity_image_cache_sweep_cb(gpointer user_data){
	if(velocity_image_cache_btree){
		VELOCITY_MSG("sweeping");
		velocity_image_cache_sweep_tmp=NULL;
		g_tree_foreach(velocity_image_cache_btree,(GTraverseFunc)velocity_image_cache_sweep_for_each_cb,NULL);
		if(velocity_image_cache_sweep_tmp){
			GSList*n=velocity_image_cache_sweep_tmp;
			while(n){
				VELOCITY_MSG("stale cache object(%s)",(char*)n->data);
				g_tree_remove(velocity_image_cache_btree,n->data);
				n=n->next;
			}
			g_slist_free(velocity_image_cache_sweep_tmp);
		}
	}
	return(TRUE);
}
static void velocity_image_init(void){
	velocity_image_git=gnome_icon_theme_new();
	gnome_icon_theme_set_allow_svg(GNOME_ICON_THEME(velocity_image_git),TRUE);
	velocity_image_cache_sweep_timer=gtk_timeout_add(VELOCITY_IMAGE_CACHE_SWEEP_INTERVAL*VELOCITY_IMAGE_CACHE_SWEEP_INTERVAL_SCALE,
		(GtkFunction)velocity_image_cache_sweep_cb,NULL);
/*	{
		GList*emblems=gnome_icon_theme_list_icons(velocity_image_git,"Emblems");
		GList*n=emblems;
		VELOCITY_MSG("emblem list");
		while(n){
			VELOCITY_MSG("emblem found(%s)",(char*)n->data);
			n=n->next;
		}
	}*/
}
gchar*velocity_image_get_uri(const gchar*image_name,guint base_size){
	char*uri=NULL;
	if(!velocity_image_git)velocity_image_init();
	gnome_icon_theme_rescan_if_needed(velocity_image_git);
	if(gnome_icon_theme_has_icon(velocity_image_git,image_name)){
		uri=gnome_icon_theme_lookup_icon(velocity_image_git,image_name,base_size,NULL,NULL);
	}else{
		uri=gnome_program_locate_file(NULL,GNOME_FILE_DOMAIN_PIXMAP,image_name,TRUE,NULL);
	}
	return(uri);
}
gchar*velocity_image_get_mime_type_uri(const gchar*mime_type,guint base_size){
	char*base;
	char*ret;
	const char*vfs=gnome_vfs_mime_get_icon(mime_type);
	if(!velocity_image_git)velocity_image_init();
	if((vfs)&&(g_path_is_absolute(vfs))){
		base=g_strdup(vfs);
	}else{
		base=gnome_icon_lookup(velocity_image_git,NULL,NULL,NULL,NULL,mime_type,GNOME_ICON_LOOKUP_FLAGS_NONE,GNOME_ICON_LOOKUP_RESULT_FLAGS_NONE);
	}
	if((!base)||(strcmp(base,"gnome-fs-regular")==0)){
		const char*tmp=gnome_vfs_mime_get_icon(mime_type);
		if(tmp)base=g_strdup(tmp);
	}
	if(base){
		if(g_path_is_absolute(base)){	/*Skip URI's that are absolute (usually user icons)*/
			ret=base;
		}else{
			ret=velocity_image_get_uri(base,base_size);
			g_free(base);
		}
	}else{
		ret=velocity_image_get_uri(VELOCITY_DEFAULT_ICON,base_size);
	}
	return(ret);
}
velocity_image_info*velocity_image_info_new_from_pixbuf(GdkPixbuf*p){
	velocity_image_info*info=g_new(velocity_image_info,1);
	info->n_channels=gdk_pixbuf_get_n_channels(p);
	info->has_alpha=gdk_pixbuf_get_has_alpha(p);
	info->bits_per_sample=gdk_pixbuf_get_bits_per_sample(p);
	info->width=gdk_pixbuf_get_width(p);
	info->height=gdk_pixbuf_get_height(p);
	info->rowstride=gdk_pixbuf_get_rowstride(p);
	return(info);
}
GdkPoint*get_scaled_aspect(const guint w,const guint h,const guint scaledw,const guint scaledh){
	GdkPoint*tmp=g_new0(GdkPoint,1);
	if(w>h){
		tmp->x=scaledw;
		tmp->y=(h*scaledh)/w;
	}else{
		tmp->x=(w*scaledw)/h;
		tmp->y=scaledh;
	}
	if(tmp->x<1)tmp->x=1;
	if(tmp->y<1)tmp->y=1;
	return(tmp);
}
static velocity_image_cachelet*velocity_image_cachelet_pixbuf_new(const gchar*uri,const guint width,const guint height,
	velocity_image_info**info_ret,velocity_image_cache_priority priority){
	velocity_image_cachelet*cl=g_new0(velocity_image_cachelet,1);
	if(cl){
		cl->data=(gpointer)gnome_gdk_pixbuf_new_from_uri(uri);
		if(cl->data){
			cl->size.x=gdk_pixbuf_get_width((GdkPixbuf*)cl->data);
			cl->size.y=gdk_pixbuf_get_height((GdkPixbuf*)cl->data);
			if(info_ret)*info_ret=velocity_image_info_new_from_pixbuf((GdkPixbuf*)cl->data);
			if(((width!=0)&&(height!=0))&&((cl->size.x>width)||(cl->size.y>height))){
				GdkPoint*sa=get_scaled_aspect(cl->size.x,cl->size.y,width,height);
				GdkPixbuf*t=(GdkPixbuf*)cl->data;
				cl->data=(gpointer)gdk_pixbuf_scale_simple(t,sa->x,sa->y,VELOCITY_IMAGE_GDK_SCALE_MODE);
				g_object_unref(t);
				g_free(sa);
			}
			cl->size.x=width;
			cl->size.y=height;
			cl->priority=priority;
			return(cl);
		}
		g_free(cl);
	}
	return(NULL);
}
static void velocity_image_cache_object_destroy(velocity_image_cache_object*co){
	if(co){
		if(co->pixbufs){
			GSList*n=co->pixbufs;
			while(n){
				if(n->data){
					velocity_image_cachelet*cl=(velocity_image_cachelet*)n->data;
					velocity_image_cachelet_free(cl);
				}
				n=n->next;
			}
			g_slist_free(co->pixbufs);
		}
		if(co->info)g_free(co->info);
		g_free(co);
		co=NULL;
	}
}
GdkPixbuf*velocity_image_cache_get(const gchar*uri,const guint width,const guint height,velocity_image_info**info_ret,velocity_image_cache_priority priority){
	GdkPixbuf*pixbuf=NULL;
	/*To help any chances of keeping myself in the sane catagory*/
	if(!velocity_image_cache_btree){
		velocity_image_cache_btree=g_tree_new_full((GCompareDataFunc)strcmp,NULL,
			(GDestroyNotify)g_free,(GDestroyNotify)velocity_image_cache_object_destroy);
	}
	if(uri){	/*only do something when it makes sense to*/
		velocity_image_cache_object*co=(velocity_image_cache_object*)g_tree_lookup(velocity_image_cache_btree,(gpointer)uri);
		velocity_image_cachelet*tcl=NULL;
		gboolean insert=FALSE;
		if(!co){
			if(priority>VELOCITY_IMAGE_CACHE_PRIORITY_NONE){
				co=g_new0(velocity_image_cache_object,1);
				insert=TRUE;	/*We use this so we know we only insert into the tree on new objects*/
			}
velocity_image_cache_force:
			tcl=velocity_image_cachelet_pixbuf_new(uri,width,height,(insert)?&co->info:NULL,priority);
			if(info_ret)*info_ret=co->info;
			if(tcl){	/*Pixbuf was loaded*/
				pixbuf=(GdkPixbuf*)tcl->data;
				g_object_ref((GdkPixbuf*)tcl->data);
				if(priority>VELOCITY_IMAGE_CACHE_PRIORITY_NONE){
					/*Add to pixbuf cache list*/
					g_object_ref((GdkPixbuf*)tcl->data);
					if((width==0)&&(height==0)){
						/*We want to keep the original size image at the end of the list for speed purposes*/
						co->pixbufs=g_slist_append(co->pixbufs,(gpointer)tcl);
					}else{
						co->pixbufs=g_slist_prepend(co->pixbufs,(gpointer)tcl);
					}
				}
				if(priority<=VELOCITY_IMAGE_CACHE_PRIORITY_NONE){
					g_object_unref((GdkPixbuf*)tcl->data);
					g_free(tcl);
				}
			}else{
				if(insert)velocity_image_cache_object_destroy(co);
				return(NULL);
			}
			/*add item to btree*/
			if(insert)g_tree_insert(velocity_image_cache_btree,(gpointer)g_strdup(uri),(gpointer)co);
		}else{
			/*Cache checking core*/
			GSList*n=co->pixbufs;
			while(n){
				if(n->data){
					velocity_image_cachelet*cl=(velocity_image_cachelet*)n->data;
					if((cl->size.x==width)&&(cl->size.y==height)){
						g_object_ref(cl->data);
						pixbuf=(GdkPixbuf*)cl->data;
//TODO:  see how useful something like this would be
/*					}else if((co->info->width==cl->size.x)&&(co->info->height==cl->size.y)&&(!n->next)){
						VELOCITY_MSG("%s already cached at full size, resize and add to cache list?",uri);
						GdkPoint*sa=get_scaled_aspect(cl->size.x,cl->size.y,width,height);
						pixbuf=(GdkPixbuf*)gdk_pixbuf_scale_simple((GdkPixbuf*)cl->data,sa->x,sa->y,VELOCITY_IMAGE_GDK_SCALE_MODE);
						g_free(sa);*/
					}
				}
				n=n->next;
			}
			if(!pixbuf){	/*If it wasn't found in the already cached data...*/
				goto velocity_image_cache_force;	/*...then we can jump to the code block above that does the loading*/
			}
			if(info_ret)*info_ret=co->info;
		}
	}
	return(pixbuf);
}
void velocity_image_cache_shutdown(void){
	if(velocity_image_cache_sweep_timer)gtk_timeout_remove(velocity_image_cache_sweep_timer);
	if(velocity_image_cache_btree)g_tree_destroy(velocity_image_cache_btree);
}
