/*
 * Velocity - A GNOME File Manager
 *
 *
 * Copyright (C) (2002) (Kyle Davis) <kyled@blueriver.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*Private functions for the Velocity module library*/

#include <stdio.h>
#include <glib.h>

#include "velocity-module.h"
#include "velocity-module-private.h"

GList*velocity_module_tracker_list=NULL;	/*Internal, used for the module tracking system*/

/*This MUST be NULL terminated or it WILL crash*/
char**_velocity_array_new(char*arg1,...){
	if(arg1){
		va_list ap;
		char**r=NULL;
		char*t;
		int count=1;
		va_start(ap,arg1);
		while((t=va_arg(ap,char*))){
			count++;
		}
		va_start(ap,arg1);	/*Start at beginning of list and do the real work*/
		r=g_new0(char*,count+1);
		count=1;
		r[0]=arg1;
		while((t=va_arg(ap,char*))){
			r[count++]=t;
		}
		r[count++]=NULL;
		va_end(ap);
		return(r);
	}
	return(NULL);
}

static velocity_module*velocity_module_load(const char*modname){
	int*iptr=NULL;
	velocity_module*vmod=g_new0(velocity_module,1);
	vmod->sohandle=dlopen(modname,RTLD_NOW);
	if(!vmod->sohandle){
		VELOCITY_ERR(dlerror());
		g_free(vmod);
		return(NULL);
	}
	/*Required functions in the modules*/
	iptr=dlsym(vmod->sohandle,"velocity_module_subsytem_version");
	vmod->init=dlsym(vmod->sohandle,"init");
	vmod->about=dlsym(vmod->sohandle,"about");
	vmod->unload=dlsym(vmod->sohandle,"unload");
	vmod->info=dlsym(vmod->sohandle,"module_info");

	/*Optional functions in the module*/
	vmod->gui_ext=dlsym(vmod->sohandle,"gui_ext");
	vmod->preferences_new=dlsym(vmod->sohandle,"preferences_new");
	vmod->preferences_free=dlsym(vmod->sohandle,"preferences_free");
	vmod->on_view_status_event=dlsym(vmod->sohandle,"on_view_status_event");
	vmod->on_view_behavior_event=dlsym(vmod->sohandle,"on_view_behavior_event");
	vmod->on_viewer_new=dlsym(vmod->sohandle,"on_viewer_new");
	vmod->on_viewer_free=dlsym(vmod->sohandle,"on_viewer_free");

	/*Check the validity of this file as a Velocity module*/
	if((!iptr)||(!vmod->init)||(!vmod->about)||(!vmod->info)||(!vmod->unload)){
		g_print("\nVELOCITY-LIB ERROR: Missing functions in '%s'.  This doesn't seem to be a valid Velocity module.\n",modname);
		dlclose(vmod->sohandle);
		g_free(vmod);
		return(NULL);
	}
	/*Might as well dereference the version ptr for you...*/
	vmod->version=(int)(*iptr);
	if(vmod->version!=VELOCITY_MODULE_SUBSYSTEM_VERSION){
		g_print("\nVELOCITY-LIB WARNING: Version mismatch in '%s'.\n"
			"\tModule system version is v%i, but v%i is required.\n"
			"\tModule not loaded.\n",modname,vmod->version,VELOCITY_MODULE_SUBSYSTEM_VERSION);
		dlclose(vmod->sohandle);
		g_free(vmod);
		return(NULL);
	}
	vmod->filename=g_strdup(modname);
	return(vmod);
}
gboolean velocity_module_is_dynamic(velocity_module*vmod){
	if((vmod->filename)&&(vmod->sohandle))return(TRUE);
	return(FALSE);
}
static void velocity_module_unload(velocity_module*vmod){
	if(vmod){
		VELOCITY_MSG("Unloading %s",vmod->info->id);
		vmod->unload();
		if(velocity_module_is_dynamic(vmod)){
			if(dlclose(vmod->sohandle)){
				VELOCITY_ERR(dlerror());
			}
			g_free(vmod->filename);
		}
		g_free(vmod);
		vmod=NULL;
	}
}
gboolean velocity_module_supports_mime_type(velocity_module*vmod,const char*mime_type){
	if((vmod)&&(vmod->info)&&(vmod->info->mime_types))return(velocity_mime_type_vector_check(vmod->info->mime_types,mime_type));
}

/*Module preferences system*/
void velocity_module_preferences_set_string(const gchar*id,const gchar*group,const gchar*key,const gchar*value){
	gchar*confdir=gnome_config_get_real_path(VELOCITY_MODULE_CONFIG_DIR);
	gchar*confstr=g_strdup_printf("=%s/%s=/%s/%s",confdir,id,group,key);
	gnome_config_set_string(confstr,value);
	gnome_config_sync();
	g_free(confstr);
	g_free(confdir);
}
gchar*velocity_module_preferences_get_string(const gchar*id,const gchar*group,const gchar*key,const gchar*defvalue){
	gchar*confdir=gnome_config_get_real_path(VELOCITY_MODULE_CONFIG_DIR);
	gchar*confstr=g_strdup_printf("=%s/%s=/%s/%s=%s",confdir,id,group,key,defvalue);
	gchar*ret=NULL;
	gnome_config_sync();
	ret=gnome_config_get_string(confstr);
	g_free(confstr);
	g_free(confdir);
	return(ret);
}
void velocity_module_preferences_set_int(const gchar*id,const gchar*group,const gchar*key,int value){
	gchar*confdir=gnome_config_get_real_path(VELOCITY_MODULE_CONFIG_DIR);
	gchar*confstr=g_strdup_printf("=%s/%s=/%s/%s",confdir,id,group,key);
	gnome_config_set_int(confstr,value);
	gnome_config_sync();
	g_free(confstr);
	g_free(confdir);
}
int velocity_module_preferences_get_int(const gchar*id,const gchar*group,const gchar*key,int defvalue){
	gchar*confdir=gnome_config_get_real_path(VELOCITY_MODULE_CONFIG_DIR);
	gchar*confstr=g_strdup_printf("=%s/%s=/%s/%s=%i",confdir,id,group,key,defvalue);
	int ret=0;
	gnome_config_sync();
	ret=gnome_config_get_int(confstr);
	g_free(confstr);
	g_free(confdir);
	return(ret);
}
void velocity_module_preferences_set_bool(const gchar*id,const gchar*group,const gchar*key,gboolean value){
	gchar*confdir=gnome_config_get_real_path(VELOCITY_MODULE_CONFIG_DIR);
	gchar*confstr=g_strdup_printf("=%s/%s=/%s/%s",confdir,id,group,key);
	gnome_config_set_bool(confstr,value);
	gnome_config_sync();
	g_free(confstr);
	g_free(confdir);
}
gboolean velocity_module_preferences_get_bool(const gchar*id,const gchar*group,const gchar*key,gboolean defvalue){
	gchar*confdir=gnome_config_get_real_path(VELOCITY_MODULE_CONFIG_DIR);
	gchar*confstr=g_strdup_printf("=%s/%s=/%s/%s=%i",confdir,id,group,key,defvalue);
	gboolean ret=FALSE;
	gnome_config_sync();
	ret=gnome_config_get_bool(confstr);
	g_free(confstr);
	g_free(confdir);
	return(ret);
}

/*Module blacklist (list of modules that should not be loaded)*/
#define VELOCITY_MODULE_INTERNAL	"internal"
#define VELOCITY_MODULE_BLACKLIST	"blacklist"
void velocity_module_blacklist_add(const gchar*id){
	velocity_module_preferences_set_bool(VELOCITY_MODULE_INTERNAL,VELOCITY_MODULE_BLACKLIST,id,TRUE);
}
void velocity_module_blacklist_remove(const gchar*id){
	velocity_module_preferences_set_bool(VELOCITY_MODULE_INTERNAL,VELOCITY_MODULE_BLACKLIST,id,FALSE);
}
gboolean velocity_module_blacklist_check(const gchar*id){
	return(velocity_module_preferences_get_bool(VELOCITY_MODULE_INTERNAL,VELOCITY_MODULE_BLACKLIST,id,FALSE));
}

/*Module tracking system*/
GList*velocity_module_tracker_get_list(void){
	return(velocity_module_tracker_list);
}
velocity_module*velocity_module_tracker_register(const char*modname){
	if(velocity_module_tracker_get(modname)){
		VELOCITY_ERR("Module %s is already loaded and can not be loaded again",modname);
	}else{
		velocity_module*vmod=velocity_module_load(modname);
		if(vmod){
			switch(vmod->init()){
				case VELOCITY_MODULE_OK:
					velocity_module_tracker_list=g_list_prepend(velocity_module_tracker_list,(gpointer)vmod);
					return(vmod);
					break;
				case VELOCITY_MODULE_ERROR:
					VELOCITY_ERR("Module %s reported an error and will be unloaded and blacklisted",vmod->info->id);
				case VELOCITY_MODULE_UNLOAD_PERSISTANT:
					velocity_module_blacklist_add(vmod->info->id);
				case VELOCITY_MODULE_UNLOAD:
				default:
					g_free(vmod);
					break;
			}
		}
	}
	return(NULL);
}
velocity_module*velocity_module_tracker_register_static(velocity_module_return(*init)(void),void(*about)(void),void(*unload)(void),
	velocity_module_info*info,velocity_module_gui_ext*gui_ext,
	velocity_view_status_event_ucb on_view_status_event,velocity_view_behavior_event_ucb on_view_behavior_event,
	velocity_module_viewer_new on_viewer_new,velocity_module_ext_callback on_viewer_free){
	if((init)&&(about)&&(unload)){
		velocity_module*vmod=g_new0(velocity_module,1);
		g_assert(vmod!=NULL);
		vmod->init=init;
		vmod->about=about;
		vmod->unload=unload;
		vmod->info=info;
		vmod->gui_ext=gui_ext;
		vmod->on_view_status_event=on_view_status_event;
		vmod->on_view_behavior_event=on_view_behavior_event;
		vmod->on_viewer_new=on_viewer_new;
		vmod->on_viewer_free=on_viewer_free;
		vmod->version=VELOCITY_MODULE_SUBSYSTEM_VERSION;	/*We'll assume static modules weren't superimposed into the binary*/
		vmod->sohandle=NULL;
		vmod->filename=NULL;
		switch(vmod->init()){
			case VELOCITY_MODULE_OK:
				velocity_module_tracker_list=g_list_prepend(velocity_module_tracker_list,(gpointer)vmod);
				return(vmod);
				break;
			case VELOCITY_MODULE_ERROR:
				VELOCITY_ERR("Static module %s reported an error and will be unloaded and blacklisted",info->id);
			case VELOCITY_MODULE_UNLOAD_PERSISTANT:
				velocity_module_blacklist_add(info->id);
			case VELOCITY_MODULE_UNLOAD:
			default:
				g_free(vmod);
				break;
		}
		return(NULL);
	}else{
		g_print("\nVELOCITY-LIB ERROR: Static module is missing functions.  This doesn't seem to be a valid Velocity module.\n");
		return(NULL);
	}
}
static gint velocity_module_tracker_find(gconstpointer a,gconstpointer b){
	velocity_module*mymod=(velocity_module*)a;
	if(velocity_module_is_dynamic(mymod))return(strcmp(mymod->filename,(const char*)b));
	return(TRUE);
}
velocity_module*velocity_module_tracker_get(const gchar*modname){
	if(velocity_module_tracker_list){
		GList*item=g_list_find_custom(velocity_module_tracker_list,(gpointer)modname,(GCompareFunc)velocity_module_tracker_find);
		if((item)&&(item->data))return((velocity_module*)item->data);
	}
	return(NULL);
}
gboolean velocity_module_tracker_unregister(velocity_module*vmod){
	if((vmod)&&(velocity_module_is_dynamic(vmod))){	/*Because I don't trust g_list_remove*/
		GList*nl=NULL;
		GList*n=velocity_module_tracker_list;
		while(n){
			if(n->data!=(gpointer)vmod)nl=g_list_append(nl,n->data);
			n=n->next;
		}
		n=velocity_module_tracker_list;
		velocity_module_tracker_list=nl;
		g_list_free(n);
		velocity_module_unload(vmod);
		return(TRUE);
	}else{
		return(FALSE);
	}
}
gboolean velocity_module_tracker_unregister_by_name(const char*modname){
	velocity_module*mymod=NULL;
	mymod=velocity_module_tracker_get(modname);
	if(mymod)return(velocity_module_tracker_unregister(mymod));
	return(FALSE);
}
void velocity_module_tracker_shutdown(void){
	/*unload all modules*/
	GList*n=NULL;
	velocity_module*vmod=NULL;
	n=velocity_module_tracker_list;
	while(n){
		vmod=(velocity_module*)n->data;
		if(vmod)velocity_module_unload(vmod);
		n=n->next;
	}
	g_list_free(velocity_module_tracker_list);
	velocity_module_tracker_list=NULL;
}
