/*
 * Velocity - A GNOME File Manager
 *
 *
 * Copyright (C) (2002) (Kyle Davis) <kyled@blueriver.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef VELOCITY_MODULE_H
#define VELOCITY_MODULE_H

#include <stdarg.h>
#include <stdio.h>
#include <dlfcn.h>	//TODO:  start using gmodule.h here.....
#include <glib.h>

#include "velocity-icon.h"
#include "velocity-vfs-object.h"
#include "velocity-view.h"
#include "velocity-window.h"
#include "velocity-preferences-window.h"

#define VELOCITY_MODULE_SUBSYSTEM_VERSION	0x0001

/*Generic type definitions for module extention callbacks*/
typedef void(*velocity_module_ext_callback)(GtkWidget*widget);

/*Type definitions for custom notebook pages*/
typedef GtkWidget*(*velocity_module_gui_ext_page_new)(velocity_vfs_obj*vfsobj);	/*Return what you want inserted into new page.  If NULL is returned, new page will not be created*/

/*Type definitions for custom view callbacks*/
typedef GtkWidget*(*velocity_module_viewer_new)(velocity_window*vwin,const char*uri);

/*each modules init() should return one of these*/
typedef enum{VELOCITY_MODULE_INVALID=0,
	VELOCITY_MODULE_OK,	/*everything is fine*/
	VELOCITY_MODULE_ERROR,	/*something went horribly wrong and that the module should be unloaded since it would be unusable anyhow*/
	VELOCITY_MODULE_UNLOAD,	/*there wasn't an error but the module should be unloaded*/
	VELOCITY_MODULE_UNLOAD_PERSISTANT,	/*unload this module and don't load it again unless the user specifies othewise*/
	VELOCITY_MODULE_NUM}velocity_module_return;

/*Version numbering*/
typedef struct{
	guint major;
	guint minor;
	guint micro;
}velocity_module_version;

/*NOTE:  ALL char** 's must be NULL terminated!*/
typedef struct{
	char*id;	/*Internal name.  Used for dependancy tracking.  (generally the module's filename as well)*/
	char*name;	/*Name/short description*/
	char*description;	/*Long description*/
	char**authors;	/*Kyle Davis  (why not...everybody deserves credit)*/
	velocity_module_version version;	/*see velocity_module_version*/
	char**deps;	/*module dependancies (by module id (case sensitive).  For modules that require other modules.*/	//TODO?
	char**mime_types;	/*related mime_types ({"internal/all",NULL}==everything we have to offer)*/
	char*url;	/*website URL (ignored if NULL)*/
}velocity_module_info;

typedef struct{
	const char*page_title;
	velocity_module_gui_ext_page_new page_new;
	velocity_module_ext_callback page_ok;	/*if ok was pressed (NULL if not needed)*/
	velocity_module_ext_callback page_cancel;	/*if cancel was pressed (NULL if not needed)*/
}velocity_module_gui_ext_page;

typedef struct{
	GList*toolbar_items;	/*to add to the toolbar*/
	GList*templates;	/*to add to the templates menu.  see velocity_static_template_new*/
	GList*pages;	/*notebook pages to add to the properties dialog (velocity_module_gui_ext_pages*'s)*/
	gpointer*tools;	/*A velocity menu factory stream of items for the "Tools" menu (not based on module's supported MIME-types)*/
	gpointer*sendto;	/*A velocity menu factory stream of items for the "Send to" menu (not based on module's supported MIME-types)*/
	gpointer*addons;	/*A velocity menu factory stream of items for the "Add-ons" menu (use when not going to be used often, but still handy)*/
	gpointer*context;	/*A velocity menu factory stream of items for the normal context menu (use when going to be used often)*/
}velocity_module_gui_ext;

typedef struct{
	/*Basics (all required)*/
	velocity_module_return(*init)(void);	/*init the module (ie, set the velocity_module_info struct)*/
	void(*about)(void);	/*an about box*/
	void(*unload)(void);	/*free module and other resources (module_info);*/
	guint version;	/*VELOCITY_MODULE_SUBSYSTEM_VERSION (for compatibility checks)*/
	velocity_module_info*info;	/*keeps info from the module's module_info(...)*/

	/*GUI hooks*/
	velocity_module_gui_ext*gui_ext;	/*GUI extentions*/
	velocity_preferences_dialog_component preferences_new;	/*a way to set preferences*/
	velocity_preferences_dialog_component_free preferences_free;	/*a func to clean up from the above*/
	//TODO:  May want to put this into a struct/union to keep things more organized
	velocity_view_status_event_ucb on_view_status_event;	/*handle view's status events*/
	velocity_view_behavior_event_ucb on_view_behavior_event;	/*handle view's behavioral events*/
	velocity_module_viewer_new on_viewer_new;	/*Returns something representing your view*/
	velocity_module_ext_callback on_viewer_free;	/*You get your view's data...you should free everything*/

	/*Internal*/
	char*filename;	/*name of module file*/
	void*sohandle;	/*internal don't use*/
}velocity_module;


/*Module preferences system*/
#define VELOCITY_MODULE_CONFIG_DIR	"Velocity_Plugins"
void velocity_module_preferences_set_string(const gchar*id,const gchar*group,const gchar*key,const gchar*value);
gchar*velocity_module_preferences_get_string(const gchar*id,const gchar*group,const gchar*key,const gchar*defvalue);
void velocity_module_preferences_set_int(const gchar*id,const gchar*group,const gchar*key,int value);
int velocity_module_preferences_get_int(const gchar*id,const gchar*group,const gchar*key,int defvalue);
void velocity_module_preferences_set_bool(const gchar*id,const gchar*group,const gchar*key,gboolean value);
gboolean velocity_module_preferences_get_bool(const gchar*id,const gchar*group,const gchar*key,gboolean defvalue);
char**_velocity_array_new(char*arg1,...);	/*Internal.  For macros*/

/*Convenience wrappers for module developers, USE THESE TO MAKE A MODULE*/
#define VELOCITY_MODULE_INFO(module_id,module_name,module_description,module_url) \
	module_info.id=#module_id; \
	module_info.name=module_name; \
	module_info.description=module_description; \
	module_info.url=module_url;
#define VELOCITY_MODULE_VERSION(vmajor,vminor,vmicro) \
	module_info.version.major=vmajor; \
	module_info.version.minor=vminor; \
	module_info.version.micro=vmicro;
#define VELOCITY_MODULE_AUTHORS(...)	module_info.authors=_velocity_array_new(__VA_ARGS__,NULL);
#define VELOCITY_MODULE_MIME_TYPES(...)	module_info.mime_types=_velocity_array_new(__VA_ARGS__,NULL);
#define VELOCITY_MODULE_DEPENDANCIES(...)	module_info.deps=_velocity_array_new(__VA_ARGS__,NULL);

#define VELOCITY_STATIC_MODULE_INFO_DEFINE(mid)	velocity_module_info mid ## _module_info
#define VELOCITY_STATIC_MODULE_GUI_DEFINE(mid)	velocity_module_gui_ext mid ## _gui_ext
#define VELOCITY_STATIC_MODULE_INFO(mid,mname,mdescription,murl) \
	mid ## _module_info.id=#mid; \
	mid ## _module_info.name=mname; \
	mid ## _module_info.description=mdescription; \
	mid ## _module_info.url=murl;
#define VELOCITY_STATIC_MODULE_VERSION(mid,vmajor,vminor,vmicro) \
	mid ## _module_info.version.major=vmajor; \
	mid ## _module_info.version.minor=vminor; \
	mid ## _module_info.version.micro=vmicro;
#define VELOCITY_STATIC_MODULE_AUTHORS(mid,...)	mid ## _module_info.authors=_velocity_array_new(__VA_ARGS__,NULL);
#define VELOCITY_STATIC_MODULE_MIME_TYPES(mid,...)	mid ## _module_info.mime_types=_velocity_array_new(__VA_ARGS__,NULL);
#define VELOCITY_STATIC_MODULE_DEPENDANCIES(mid,...)	mid ## _module_info.deps=_velocity_array_new(__VA_ARGS__,NULL);
#define VELOCITY_STATIC_MODULE_INFO_VAR(mid) mid ## _module_info
#define VELOCITY_STATIC_MODULE_GUI_VAR(mid) mid ## _gui_ext

#endif
