/*
 * Velocity - A GNOME File Manager
 *
 *
 * Copyright (C) (2002) (Kyle Davis) <kyled@blueriver.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <ctype.h>
#include "velocity-view.h"
#include "velocity-preferences.h"
#include "velocity-icons.h"
#include "velocity-thumbnail.h"
#include "velocity-ditem.h"
#include "velocity-dnd.h"

velocity_clipboard*vclipboard=NULL;	/*Global clipboard*/
static GtkWidget*velocity_view_tooltip=NULL;	/*Using a global because there should only be one tooltip throughout all of the vview objects on screen*/
velocity_vfs_obj*velocity_view_last;	/*Last item that recieved an event.  Same reason for global here too.*/
static gboolean velocity_view_clicked=FALSE;	/*Used for DnD*/
static GdkPoint velocity_view_click;	/*Used for DnD*/

static void velocity_view_free_cb(GtkObject*obj,gpointer user_data);
static gint velocity_view_canvas_cb(GnomeCanvas*canvas,GdkEvent*event,gpointer user_data);
static gint velocity_view_icon_cb(GnomeCanvasItem*item,GdkEvent*event,gpointer user_data);
static gint velocity_view_animation_cb(gpointer user_data);
static void velocity_view_loading_cb(GnomeVFSAsyncHandle*handle,GnomeVFSResult result,GList*list,guint entries_read,gpointer user_data);
static gint velocity_view_dequeue_cb(velocity_view*vview);
static void velocity_view_validated_cb(GnomeVFSAsyncHandle*handle,GList*results,gpointer user_data);
static void velocity_view_goto(velocity_view*vview,const char*location);
static void velocity_view_validate(velocity_view*vview,const char*location);
static void velocity_view_tooltip_off(velocity_view*vview);

GdkPixbuf*velocity_view_get_mime_icon(velocity_view*vview,velocity_vfs_obj*vfsobj,guint width,guint height){
	GdkPixbuf*p=NULL;
	if(vview->on_behavior_event)p=vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_OBJECT_GET_ICON,NULL,vfsobj,vview->user_data);
	if(!p){
		char*uri=velocity_vfs_get_uri_unescaped(vfsobj);
		if(uri){
			char*puri=velocity_vfs_get_mime_icon_uri(vfsobj->info->mime_type,uri,MAX(width,height));
			p=velocity_image_cache_get(puri,width,height,NULL,VELOCITY_IMAGE_CACHE_PRIORITY_HIGH);
			g_free(puri);
			g_free(uri);
		}
	}
	if(vfsobj->info->name[0]=='.'){	/*TODO:  As simple as this is, it should probably still be in a function somewhere*/
		if(p){	/*This causes hidden files to be half transparent*/
			GdkPixbuf*t=p;
			int w=gdk_pixbuf_get_width(t);
			int h=gdk_pixbuf_get_height(t);
			p=gdk_pixbuf_new(gdk_pixbuf_get_colorspace(t),FALSE,gdk_pixbuf_get_bits_per_sample(t),w,h);
			gdk_pixbuf_fill(p,0xFFFFFF00);
			gdk_pixbuf_composite(t,p,0,0,w,h,0.0,0.0,1.0,1.0,GDK_INTERP_NEAREST,128);
			g_object_unref(t);
		}
	}
	return(p);
}
static void velocity_view_drag_begin_cb(GtkWidget*widget,GdkDragContext*context,gpointer user_data){
	velocity_view*vview=(velocity_view*)user_data;
	if(velocity_clipboard_has_items(vview->selection)){
		if(velocity_clipboard_has_multiple(vview->selection)){
			//TODO:  We need an icon for multiple files w/ a non common mime type....probably a couple icons stacked...seems logical...
		}else{
			velocity_vfs_obj*vfsobj=(velocity_vfs_obj*)vview->selection->items->data;
			GdkPixbuf*pb=velocity_view_get_mime_icon(vview,vfsobj,
				(vview->prefs->profile->icon_size.x*GNOME_CANVAS(vview->canvas)->pixels_per_unit),
				(vview->prefs->profile->icon_size.y*GNOME_CANVAS(vview->canvas)->pixels_per_unit));
			if(pb){
				GdkPixmap*p=NULL;
				GdkBitmap*m=NULL;
				gdk_pixbuf_render_pixmap_and_mask(pb,&p,&m,128);
				if((p)&&(m)){
					gtk_drag_set_icon_pixmap(context,gtk_widget_get_colormap(widget),p,m,0,0);
					g_object_unref(p);
					g_object_unref(m);
				}
				g_object_unref(pb);
			}
		}
	}
}
static void velocity_view_drag_get_cb(GtkWidget*widget,GdkDragContext*context,GtkSelectionData*data,guint info,guint32 time,gpointer user_data){
	if(data){
		velocity_view*vview=(velocity_view*)user_data;
		GString*uris=g_string_new("");
		GList*n=vview->selection->items;
		switch(info){
			case VELOCITY_DND_URI_LIST:
			case VELOCITY_DND_URI:
				while(n){
					if(n->data){
						velocity_vfs_obj*vfsobj=(velocity_vfs_obj*)n->data;
						gchar*uri=velocity_vfs_get_uri_unescaped(vfsobj);
						gchar*li=g_strdup_printf("%s\r\n",uri);
						g_string_append(uris,li);
						g_free(li);
						g_free(uri);
					}
					n=n->next;
				}
				gtk_selection_data_set(data,data->target,8,uris->str,strlen(uris->str));
				break;
			default:
				g_assert_not_reached();
		}
		g_string_free(uris,TRUE);
	}
}
static gboolean velocity_view_drag_motion_cb(GtkWidget*widget,GdkDragContext*context,gint x,gint y,guint time,gpointer user_data){
	velocity_view*vview=(velocity_view*)user_data;
	if((vview)&&(vview->lasso))velocity_view_show_lasso(vview,vview->lasso->fx1,vview->lasso->fy1,x,y);
	return(FALSE);
}
static gboolean velocity_view_drag_received_cb(GtkWidget*widget,GdkDragContext*context,gint x,gint y,GtkSelectionData*data,guint info,guint32 time,gpointer user_data){
	velocity_view*vview=(velocity_view*)user_data;
	if((data)&&(data->length>0)){
		GList*uris=gnome_vfs_uri_list_parse(data->data);
		GnomeCanvasItem*item=NULL;
		gboolean s=FALSE;
		gdouble wx;
		gdouble wy;
		gnome_canvas_window_to_world(GNOME_CANVAS(widget),x,y,&wx,&wy);
		item=gnome_canvas_get_item_at(GNOME_CANVAS(widget),wx,wy);
		wx+=GTK_RANGE(GTK_SCROLLED_WINDOW(vview->widget)->hscrollbar)->adjustment->value;
		wy+=GTK_RANGE(GTK_SCROLLED_WINDOW(vview->widget)->vscrollbar)->adjustment->value;
		if((item)&&(gtk_drag_get_source_widget(context)==widget)){
			GList*n=vview->selection->items;
			while(n){
				velocity_vfs_obj*vfsobj=(velocity_vfs_obj*)n->data;
				if(velocity_icon_is_child(vfsobj->icon,GNOME_CANVAS_ITEM(item))){
					s=TRUE;
					break;
				}
				n=n->next;
			}
		}
		if((!item)||(s)){
			if(!uris){
				g_warning("No Drag & Drop URI's");
				gtk_drag_finish(context,FALSE,FALSE,time);
			}else{
				if(gtk_drag_get_source_widget(context)==widget){	/*If this is the same vview that we came from just move icons around*/
					GList*n=vview->selection->items;
					gdouble xi=0;
					gdouble yi=0;
					while(n){
						if(n->data){
							velocity_vfs_obj*vfsobj=(velocity_vfs_obj*)n->data;
							if(n==vview->selection->items){
								xi=wx-vfsobj->icon->x1-1;
								yi=wy-vfsobj->icon->y1-1;
							}
							velocity_icon_move(vfsobj->icon,xi,yi);
						}
						n=n->next;
					}
				}else{
					//TODO:  GENERAL SANITY CHECKING...
					//TODO:  CHECK IF THIS IS A SAFE LOCATION (eg. we're allowed to drag here, this is a directory, etc.)!!!!
					//TODO:  ENSURE THAT NONE OF THE SPECIAL DIRS ARE BEING MOVED (OR IF THEY ARE THAT THEY ARE ALLOWED TO BE MOVED)
					//TODO:  ALSO - MAKE SURE THAT YOU *COPY* RATHER THAN MOVE FOR CERTIAN URI'S
					GList*n=uris;
					GList*srcuris=NULL;
					GList*desturis=NULL;
					while(n){
						GnomeVFSURI*guri=(GnomeVFSURI*)n->data;
						char*curi=gnome_vfs_uri_to_string(guri,GNOME_VFS_URI_HIDE_NONE);
						if(curi){
							char*file=velocity_vfs_uri_get_tail(curi);
							char*uri=g_strconcat(vview->clocation,file,NULL);
							VELOCITY_MSG("dnd_recv(%s)",curi);
							srcuris=g_list_prepend(srcuris,gnome_vfs_uri_dup(guri));
							desturis=g_list_prepend(desturis,gnome_vfs_uri_new(uri));
							g_free(file);
							g_free(curi);
						}
						n=n->next;
					}
					velocity_vfs_xfer(GTK_WINDOW(widget),srcuris,desturis,VELOCITY_VFS_XFER_MOVE,TRUE,NULL,NULL);
				}
				gnome_vfs_uri_list_free(uris);
				gtk_drag_finish(context,TRUE,FALSE,time);
				return(TRUE);
			}
		}else{
			//TODO:  handle drop onto an object...
		}
		gnome_vfs_uri_list_free(uris);
	}else{
		VELOCITY_ERR("Invalid drag & drop or no data.");
	}
	return(FALSE);
}
velocity_view*velocity_view_new(GtkWindow*parent,gboolean scrollbars){
	velocity_view*vview=g_new0(velocity_view,1);
	g_assert(vview!=NULL);
	vview->prefs=g_new0(velocity_view_preferences,1);
	g_assert(vview->prefs!=NULL);
	vview->parent=parent;
	/*Initialize globals if needed*/
	if(!vclipboard)vclipboard=velocity_clipboard_new();

	vview->selection=velocity_clipboard_new();
	vview->border.top=vview->border.right=vview->border.bottom=vview->border.left=0;
	vview->widget=gtk_scrolled_window_new(NULL,NULL);
	g_signal_connect(vview->widget,"destroy",G_CALLBACK(velocity_view_free_cb),vview);
	if(scrollbars){
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(vview->widget),GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
	}else{
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(vview->widget),GTK_POLICY_NEVER,GTK_POLICY_NEVER);
	}
	gtk_widget_show(vview->widget);
	return(vview);
}
void gtk_widget_set_background_color(GtkWidget*w,GdkColor*color){
	/*try to alloc color*/
	if(gdk_color_alloc(gdk_colormap_get_system(),color)){
		/*set new style*/
		GtkStyle*style=gtk_widget_get_style(GTK_WIDGET(w));
		style->bg[GTK_STATE_NORMAL]=*color;
	}
}
void velocity_view_set_background_color(velocity_view*vview,GdkColor*color){
	vview->prefs->background.red=color->red;
	vview->prefs->background.green=color->green;
	vview->prefs->background.blue=color->blue;
	velocity_view_ensure_canvas(vview);
	gtk_widget_set_background_color(GTK_WIDGET(vview->canvas),color);
}
void velocity_view_append_unsorted(velocity_view*vview,velocity_vfs_obj*vfsobj){
	/*Update list size*/
	vview->total.objects++;
	if(velocity_vfs_is_directory(vfsobj)){
		vview->total.dirs++;
	}else{
		vview->total.files++;
	}
	vfsobj->data=vview;
	vview->files=g_list_prepend(vview->files,vfsobj);
	/*This is because there is a test later where this is used to free everybody, otherwise we'd have a big crash*/
	if(!vview->prefs->dirs_first)vview->dirs=vview->files;
	vview->qn=vview->files;
	if(!vview->dequeue_handle)vview->dequeue_handle=gtk_idle_add((GtkFunction)velocity_view_dequeue_cb,(gpointer)vview);
}
void velocity_view_dequeue(velocity_view*vview){
	/*Dequeue the new files in the list...*/
	if(vview->on_status_event)vview->on_status_event(vview,VELOCITY_VIEW_STATUS_EVENT_LOADING,vview->clocation,GNOME_VFS_OK,vview->user_data);
	vview->dequeue_handle=gtk_idle_add((GtkFunction)velocity_view_dequeue_cb,(gpointer)vview);
}
void velocity_view_dequeue_1(velocity_view*vview){
	velocity_view_dequeue_cb(vview);
}
/*Sorting routines*/
static gint velocity_view_sort_by_name(velocity_vfs_obj*one,velocity_vfs_obj*two){
	return((gint)strcasecmp(one->info->name,two->info->name));
}
static gint velocity_view_sort_by_type(velocity_vfs_obj*one,velocity_vfs_obj*two){
	gint ret=(gint)strcasecmp(one->info->mime_type,two->info->mime_type);
	if(ret==0)ret=velocity_view_sort_by_name(one,two);
	return(ret);
}
static gint velocity_view_sort_by_size(velocity_vfs_obj*one,velocity_vfs_obj*two){
	if(one->info->size==two->info->size)return(velocity_view_sort_by_name(one,two));
	return((one->info->size<two->info->size)?-1:1);
}
static gint velocity_view_sort_by_accessed(velocity_vfs_obj*one,velocity_vfs_obj*two){
	return(one->info->atime<two->info->atime);
}
static gint velocity_view_sort_by_modified(velocity_vfs_obj*one,velocity_vfs_obj*two){
	return(one->info->mtime<two->info->mtime);
}
static gint velocity_view_sort_by_created(velocity_vfs_obj*one,velocity_vfs_obj*two){
	return(one->info->ctime<two->info->ctime);
}
GCompareFunc velocity_view_sort_table[]={NULL,
	(GCompareFunc)velocity_view_sort_by_name,
	(GCompareFunc)velocity_view_sort_by_type,
	(GCompareFunc)velocity_view_sort_by_size,
	(GCompareFunc)velocity_view_sort_by_accessed,
	(GCompareFunc)velocity_view_sort_by_modified,
	(GCompareFunc)velocity_view_sort_by_created,
	NULL};
static gint velocity_view_resize_idle_cb(gpointer user_data){
	velocity_view*vview=(velocity_view*)user_data;
	if(vview){
		if((vview->widget)&&(vview->canvas)){
			guint w=vview->widget->allocation.width;
			guint h=vview->widget->allocation.height;
			double ppu=GNOME_CANVAS(vview->canvas)->pixels_per_unit;
			if((vview->prefs->profile->draw_start==GTK_ANCHOR_NW)&&(vview->prefs->profile->draw_direction==GTK_ANCHOR_EAST)){
				gnome_canvas_set_scroll_region(GNOME_CANVAS(vview->canvas),0.0,0.0,(w-25)*ppu,(vview->icon.y+vview->prefs->profile->icon_spacing.y)*ppu);
			}else if((vview->prefs->profile->draw_start==GTK_ANCHOR_NW)&&(vview->prefs->profile->draw_direction==GTK_ANCHOR_SOUTH)){
				gnome_canvas_set_scroll_region(GNOME_CANVAS(vview->canvas),0.0,0.0,(vview->icon.x+vview->prefs->profile->icon_spacing.x)*ppu,(h-25)*ppu);
			}
			if(!vview->dequeue_handle)velocity_view_soft_refresh(vview,FALSE);
		}
		vview->refresh_handle=(guint)NULL;
	}
	return(FALSE);
}
static void velocity_view_resize_cb(GtkWidget*w,GtkAllocation*alloc,gpointer user_data){
	/*This will put a request in for a refresh/rearrange when GTK is idle*/
	velocity_view*vview=(velocity_view*)user_data;
	if(vview){
		if(vview->refresh_handle)gtk_idle_remove(vview->refresh_handle);
		vview->refresh_handle=gtk_idle_add_priority(GTK_PRIORITY_RESIZE,(GtkFunction)velocity_view_resize_idle_cb,(gpointer)vview);
	}
}
static gint velocity_view_canvas_cb(GnomeCanvas*canvas,GdkEvent*event,gpointer user_data){
	velocity_view*vview=(velocity_view*)user_data;
	GnomeCanvasItem*i=NULL;
	double item_x;
	double item_y;
	if((!canvas)||(!event)||(!user_data))return(FALSE);
	switch(event->type){
		case GDK_SCROLL:
			if(vview->prefs->profile->draw_direction==GTK_ANCHOR_SOUTH){	/*This is a hack for scrolling horizontally when it makes sense...*/
				if(event->scroll.direction==GDK_SCROLL_DOWN){
					event->scroll.direction=GDK_SCROLL_RIGHT;
				}else if(event->scroll.direction==GDK_SCROLL_UP){
					event->scroll.direction=GDK_SCROLL_LEFT;
				}
			}
			break;
		case GDK_LEAVE_NOTIFY:
		case GDK_FOCUS_CHANGE:
			if(velocity_view_last){
				if(velocity_view_last->icon)velocity_icon_unprelight(velocity_view_last->icon);
				velocity_view_tooltip_off(vview);
				if(vview->on_behavior_event)vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_MOUSE_OUT,event,velocity_view_last,vview->user_data);
				velocity_vfs_obj_free(velocity_view_last);
				velocity_view_last=NULL;
			}
			velocity_view_tooltip_off(vview);
			break;
		case GDK_MOTION_NOTIFY:
			gnome_canvas_window_to_world(canvas,event->motion.x,event->motion.y,&item_x,&item_y);
			i=gnome_canvas_get_item_at(canvas,item_x,item_y);
			if(!i){
				velocity_view_tooltip_off(vview);
				if((velocity_view_last!=NULL)&&(velocity_view_last->icon!=NULL)){
					if(!velocity_view_last->icon->highlighted){	/*For all practical purposes, for us highlighted means in the selection queue*/
						velocity_icon_unprelight(velocity_view_last->icon);
					}
					velocity_view_tooltip_off(vview);
					if(vview->on_behavior_event)vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_MOUSE_OUT,event,velocity_view_last,vview->user_data);
					velocity_vfs_obj_free(velocity_view_last);
					velocity_view_last=NULL;
				}
			}
			if(velocity_view_clicked){
				if(vview->lasso){
					GList*n=vview->dirs;
					guint li=0;
					int sx;
					int sy;
					/*Scroll canvas when lasso is near edges*/
					gnome_canvas_get_scroll_offsets(GNOME_CANVAS(vview->canvas),&sx,&sy);
					if((int)item_x-(int)sx<20)sx-=8;
					if((int)item_x-(int)sx>vview->widget->allocation.width-20)sx+=8;
					if((int)item_y-(int)sy<20)sy-=8;
					if((int)item_y-(int)sy>vview->widget->allocation.height-20)sy+=8;
					gnome_canvas_scroll_to(GNOME_CANVAS(vview->canvas),sx,sy);
					velocity_view_show_lasso(vview,velocity_view_click.x,velocity_view_click.y,(gfloat)item_x,(gfloat)item_y);
					//TODO: Figure out a faster way of doing this....
					velocity_clipboard_flush(vview->selection);
velocity_view_canvas_selection_jmp:
					while(n){
						if(n->data){
							velocity_vfs_obj*vfsobj=(velocity_vfs_obj*)n->data;
							if(vfsobj->icon){
								GdkPoint mid;
								mid.x=(int)vfsobj->icon->x1+(vfsobj->icon->size.x>>1);
								mid.y=(int)vfsobj->icon->y1+(vfsobj->icon->size.y>>1);
								if((mid.x>=vview->lasso->fx1)&&(mid.x<=vview->lasso->fx2)&&
									(mid.y>=vview->lasso->fy1)&&(mid.y<=vview->lasso->fy2)){
									velocity_clipboard_prepend_file(vview->selection,vfsobj);
									velocity_icon_highlight(vfsobj->icon);
								}else{
									if((vfsobj->icon->highlighted)||(vfsobj->icon->prelighted)){
										velocity_icon_unhighlight(vfsobj->icon);
									}
								}
							}
						}
						n=n->next;
					}
					li++;
					if((li<2)&&(vview->dirs!=vview->files)){
						n=vview->files;
						goto velocity_view_canvas_selection_jmp;
					}
				}else{
					if(gtk_drag_check_threshold(vview->canvas,velocity_view_click.x,velocity_view_click.y,item_x,item_y)){
						i=gnome_canvas_get_item_at(GNOME_CANVAS(vview->canvas),(double)velocity_view_click.x,(double)velocity_view_click.y);
						if(!i){
							if(vview->prefs->can_lasso){
								velocity_view_show_lasso(vview,(gfloat)velocity_view_click.x,(gfloat)velocity_view_click.y,(gfloat)item_x,(gfloat)item_y);
								gtk_grab_add(GTK_WIDGET(vview->widget));
							}
						}else{
							if(vview->prefs->can_drag){
								GtkTargetList*tl=gtk_target_list_new(velocity_dnd_targets,VELOCITY_DND_TARGETS_NUM);
								gtk_drag_begin(GTK_WIDGET(vview->canvas),tl,GDK_ACTION_DEFAULT|GDK_ACTION_COPY|GDK_ACTION_LINK,1,event);
								gtk_target_list_unref(tl);
							}
						}
					}
				}
			}
			break;
		case GDK_BUTTON_PRESS:
			/*Set a focus somewhere in the canvas so we can grab keyboard events*/
			gnome_canvas_window_to_world(GNOME_CANVAS(vview->canvas),event->button.x,event->button.y,&item_x,&item_y);
			gnome_canvas_item_grab_focus(GNOME_CANVAS_ITEM(GNOME_CANVAS_GROUP(vview->canvas_group)));
			velocity_view_clicked=(event->button.button==1);
			velocity_view_click.x=(gint)item_x;
			velocity_view_click.y=(gint)item_y;
			i=gnome_canvas_get_item_at(GNOME_CANVAS(vview->canvas),item_x,item_y);
			if(!i){
				switch(event->button.button){
					case 1:	/*Left click*/
						//TODO: This still deselects on a shift+lasso drag because it deselects EVERYTHING and then selects things in it..fix!!!
						if(!((event->button.state&GDK_SHIFT_MASK)||(event->button.state&GDK_CONTROL_MASK)))velocity_view_clear_selection(vview);
						break;
					case 3: /*Right click*/
						velocity_view_clear_selection(vview);
						if(vview->on_behavior_event)vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_BACKGROUND_CONTEXT,event,NULL,vview->user_data);
						break;
				}
				if(velocity_clipboard_has_items(vview->selection))velocity_view_clear_selection(vview);
			}
			break;
		case GDK_BUTTON_RELEASE:
			velocity_view_clicked=FALSE;
			if((vview->prefs->can_lasso)&&(vview->lasso)){
				velocity_view_hide_lasso(vview);
				gtk_grab_remove(gtk_grab_get_current());
			}
			break;
		case GDK_KEY_PRESS:
			if(vview->on_behavior_event)if(vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_KEYBOARD,event,NULL,vview->user_data))break;
			switch(event->key.keyval){
				case GDK_Page_Up:
					{
						GtkAdjustment*adj=gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(vview->widget));
						gtk_adjustment_set_value(adj,adj->value-adj->page_size);
						gtk_scrolled_window_set_vadjustment(GTK_SCROLLED_WINDOW(vview->widget),adj);
					}
					break;
				case GDK_Page_Down:
					{
						GtkAdjustment*adj=gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(vview->widget));
						gtk_adjustment_set_value(adj,adj->value+adj->page_size);
						gtk_scrolled_window_set_vadjustment(GTK_SCROLLED_WINDOW(vview->widget),adj);
					}
					break;
				case GDK_Escape:
					if(vview->prefs->can_cancel)velocity_view_stop(vview);
					break;
				default:
					if(((event->key.keyval>=GDK_a)&&(event->key.keyval<=GDK_z))||
							((event->key.keyval>=GDK_A)&&(event->key.keyval<=GDK_Z))||
							((event->key.keyval>=GDK_0)&&(event->key.keyval<=GDK_9))){
						char c=tolower(event->key.keyval);
						GList*n;
						velocity_vfs_obj*vfsobj=NULL;
						//TODO:  do this differently:  s=start GList*,n=search GList* (s->next), loop through all lists, if (n==s) then break
						//TODO:  scrollto position of icon that gets selected
						gboolean filelist=FALSE;
						if((!velocity_clipboard_has_items(vview->selection))||(velocity_clipboard_has_multiple(vview->selection))){
							n=vview->dirs;
						}else{
							vfsobj=(velocity_vfs_obj*)vview->selection->items->data;
							if(tolower(vfsobj->info->name[0])!=c){
								velocity_view_clear_selection(vview);
								n=vview->dirs;
							}else{
								n=g_list_find(vview->dirs,(gpointer)vfsobj);
								if(!n){
									n=g_list_find(vview->files,(gpointer)vfsobj);
									filelist=TRUE;
								}
								if(n)n=n->next; /*We don't want the same as last time now do we*/
							}
						}
						while(n){
							vfsobj=(velocity_vfs_obj*)n->data;
							if(tolower(vfsobj->info->name[0])==tolower(c)){
								velocity_view_clear_selection(vview);
								velocity_clipboard_prepend_file(vview->selection,vfsobj);
								velocity_icon_highlight(vfsobj->icon);
								break;
							}
							n=n->next;
							if((!n)&&(!filelist)&&(vview->dirs!=vview->files)){
								n=vview->files;
								filelist=TRUE;
							}
						}
					}
					break;
			}
			break;
		default:
			break;
	}
	return(FALSE);
}
static void velocity_view_tooltip_destroy(){
	if(velocity_view_tooltip){
		gtk_object_destroy(GTK_OBJECT(velocity_view_tooltip));
		velocity_view_tooltip=NULL;
	}
}
static void velocity_view_tooltip_off(velocity_view*vview){
	if(vview->tooltip_handle){
		gtk_timeout_remove(vview->tooltip_handle);
		vview->tooltip_handle=(guint)NULL;
	}
	velocity_view_tooltip_destroy();
}
static void velocity_view_tooltip_paint_cb(GtkWidget*widget,GdkEventExpose*event,gpointer user_data){
	gtk_paint_flat_box(velocity_view_tooltip->style,velocity_view_tooltip->window,GTK_STATE_NORMAL,GTK_SHADOW_OUT,
		NULL,velocity_view_tooltip,"tooltip",0,0,-1,-1);
	gtk_paint_layout(velocity_view_tooltip->style,velocity_view_tooltip->window,GTK_STATE_NORMAL,TRUE,
		NULL,velocity_view_tooltip,"tooltip",8,8,PANGO_LAYOUT(user_data));
	g_object_unref(G_OBJECT(PANGO_LAYOUT(user_data)));
}
static gint velocity_view_tooltip_cb(gpointer user_data){
	velocity_view*vview=(velocity_view*)user_data;
	//TODO:  This still needs redone to support pixbufs and text shouldn't be a big string, but rather seperate text items...
	velocity_view_tooltip_off(vview);
	if((velocity_view_last)&&(velocity_view_last->icon)){	/*Just in case our item got removed since then*/
		int x,y;
		gtk_widget_get_pointer(vview->widget,&x,&y);
		if((x>0)&&(y>0)){
			GList*n=velocity_tooltip_items(velocity_view_last);
			GnomeCanvasRect*rect;
			GnomeCanvasText*text;
			PangoLayout*layout;
			char*str=NULL;
			gdk_window_get_pointer(NULL,&x,&y,NULL);
			while(n){
				if(n->data){
					velocity_tooltip_item*i=(velocity_tooltip_item*)n->data;
					switch(i->type){
						case VELOCITY_TOOLTIP_ITEM_TYPE_INT:
							if(i->data){
								char*line=g_strdup_printf("<b>%s</b>:  %i\n",i->desc,GPOINTER_TO_INT(i->data));
								if(str){
									char*tmp=str;
									str=g_strconcat(tmp,line,NULL);
									g_free(tmp);
								}else{
									str=g_strdup(line);
								}
								g_free(line);
							}
							break;
						case VELOCITY_TOOLTIP_ITEM_TYPE_TEXT:
							if(i->data){
								velocity_tooltip_item_text_cb cb=(velocity_tooltip_item_text_cb)i->data;
								char*txt=cb(velocity_view_last);
								if(txt){
									char*line=g_strdup_printf("<b>%s</b>:  %s\n",i->desc,txt);
									if(str){
										char*tmp=str;
										str=g_strconcat(tmp,line,NULL);
										g_free(tmp);
									}else{
										str=g_strdup(line);
									}
									g_free(line);
									g_free(txt);
								}
							}
							break;
						default:
							break;
					}
				}
				n=n->next;
			}
			velocity_view_tooltip=gtk_window_new(GTK_WINDOW_POPUP);
			gtk_widget_set_app_paintable(velocity_view_tooltip,TRUE);
			gtk_window_set_resizable(GTK_WINDOW(velocity_view_tooltip),FALSE);
			layout=gtk_widget_create_pango_layout(velocity_view_tooltip,NULL);
			gtk_widget_set_name(velocity_view_tooltip,"gtk-tooltips");
			g_signal_connect(velocity_view_tooltip,"expose_event",G_CALLBACK(velocity_view_tooltip_paint_cb),layout);
			gtk_widget_ensure_style(velocity_view_tooltip);
			pango_layout_set_markup(layout,str,strlen(str));
			gtk_window_move(GTK_WINDOW(velocity_view_tooltip),x+1,y+1);
			pango_layout_get_pixel_size(layout,&x,&y);
			gtk_widget_set_size_request(velocity_view_tooltip,x+16,y+16);
			gtk_widget_show(velocity_view_tooltip);
			g_free(str);
		}
		vview->tooltip_handle=(guint)NULL;
	}
	return(FALSE);
}
static void velocity_view_force_info_update(velocity_view*vview,velocity_vfs_obj*vfsobj){
	/*Update the object's file information (size, permissions, etc.) on click*/
	//TODO:  This should be an option for local & networked VFS's
	if(velocity_vfs_is_local(vfsobj)){
		char*uri=velocity_vfs_get_uri(vfsobj);
		if(velocity_vfs_is_symlink(vfsobj)){
			guint l=strlen(uri)-1;
			if(uri[l]=='/')uri[l]=(char)NULL;	/*I hate GnomeVFS right now....*/
		}
		if(gnome_vfs_get_file_info(uri,vfsobj->info,GNOME_VFS_FILE_INFO_DEFAULT|GNOME_VFS_FILE_INFO_GET_MIME_TYPE|GNOME_VFS_FILE_INFO_FOLLOW_LINKS)==GNOME_VFS_OK){
			VELOCITY_MSG("Updated file info for %s (%s)",uri,vfsobj->info->mime_type);
		}else{
			VELOCITY_MSG("Unable to update file info for %s",uri);
			velocity_clipboard_flush(vview->selection);
			if(vfsobj->icon)velocity_icon_hide(vfsobj->icon);
		}
		g_free(uri);
	}
}
static gint velocity_view_icon_cb(GnomeCanvasItem*item,GdkEvent*event,gpointer user_data){
	velocity_view*vview=NULL;
	velocity_vfs_obj*vfsobj=(velocity_vfs_obj*)user_data;
	static double x;
	static double y;
	static GdkEvent*last_event=NULL;
	double item_x=0;
	double item_y=0;
	gboolean retval=FALSE;
	if((!vfsobj)||(!vfsobj->icon)||(!vfsobj->info)||(!vfsobj->info->name))return(TRUE);
	vview=(velocity_view*)vfsobj->data;
	if(!last_event)last_event=gdk_event_copy(event);
	switch(event->type){
		case GDK_BUTTON_PRESS:
			item_x=event->button.x;
			item_y=event->button.y;
			/*Let's just stick to the buttons we care about*/
			if((event->button.button!=1)&&(event->button.button!=3))break;
			if(!velocity_view_last)velocity_view_last=velocity_vfs_obj_dup(vfsobj);
			/*Set a focus somewhere in the canvas so we can grab keyboard events*/
			gnome_canvas_item_grab_focus(GNOME_CANVAS_ITEM(GNOME_CANVAS_GROUP(vview->canvas_group)));
			if(event->button.button==1)velocity_view_force_info_update(vview,vfsobj);
			if((event->button.state&GDK_SHIFT_MASK)||(event->button.state&GDK_CONTROL_MASK)){
				if(velocity_clipboard_has_items(vview->selection)){
					GList*n=g_list_find(vview->selection->items,vfsobj);
					if(n){	/*If its already selected, remove it from the selection list*/
						velocity_view_set_icon_caption(vview,vfsobj,TRUE);
						velocity_icon_unprelight(vfsobj->icon);
						velocity_icon_unhighlight(vfsobj->icon);
						velocity_clipboard_remove_file(vview->selection,vfsobj);
						break;
					}
				}
			}else{
				if(!vfsobj->icon->highlighted)velocity_view_clear_selection(vview);
			}
			if(!vfsobj->icon->highlighted){
				velocity_view_set_icon_caption(vview,vfsobj,FALSE);
				velocity_icon_highlight(vfsobj->icon);
				gnome_canvas_item_raise_to_top(item);
				velocity_clipboard_set_details(vview->selection,vview->clocation,VELOCITY_CLIPBOARD_COPY);
				velocity_clipboard_prepend_file(vview->selection,(gpointer)vfsobj);
			}
			velocity_view_tooltip_off(vview);
			if(velocity_view_last)velocity_vfs_obj_free(velocity_view_last);
			velocity_view_last=velocity_vfs_obj_dup(vfsobj);
			switch(event->button.button){
				case 1: /*Left click*/
					if(vview->on_behavior_event)vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_CLICK,event,vfsobj,vview->user_data);
					break;
				case 3: /*Right click*/
					if(vview->on_behavior_event)vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_CONTEXT,event,vfsobj,vview->user_data);
					break;
				default:
					break;
			}
			retval=TRUE;
			break;
		case GDK_BUTTON_RELEASE:
		case GDK_2BUTTON_PRESS:
			item_x=event->button.x;
			item_y=event->button.y;
			/*Let's just stick to the buttons we care about*/
			if((event->button.button!=1)&&(event->button.button!=3))break;
			if(((vview->prefs->single_click)||(event->type==GDK_2BUTTON_PRESS))
				&&(!((event->button.state&GDK_SHIFT_MASK)||(event->button.state&GDK_CONTROL_MASK)))){
				switch(event->button.button){
					case 1:
						velocity_view_activate(vview);
						break;
					default:
						break;
				}
			}
			retval=TRUE;
			break;
      case GDK_MOTION_NOTIFY:
			item_x=event->motion.x;
			item_y=event->motion.y;
			if(velocity_view_last==vfsobj)break;
			if(velocity_view_last){
				if(velocity_view_last->icon){
					/*If not selected; if just prelit, then unprelight*/
					if(!velocity_view_last->icon->highlighted)velocity_icon_unprelight(velocity_view_last->icon);
					velocity_view_tooltip_off(vview);
					if(vview->on_behavior_event)vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_MOUSE_OUT,event,velocity_view_last,vview->user_data);
				}
				velocity_vfs_obj_free(velocity_view_last);
				velocity_view_last=NULL;
			}
			velocity_icon_prelight(vfsobj->icon);
			if(vview->prefs->tooltip_delay>0){
				vview->tooltip_handle=gtk_timeout_add(vview->prefs->tooltip_delay,(GtkFunction)velocity_view_tooltip_cb,(gpointer)vview);
			}
			if(vview->on_behavior_event)vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_MOUSE_OVER,event,vfsobj,vview->user_data);
			velocity_view_last=velocity_vfs_obj_dup(vfsobj);
			retval=TRUE;
			break;
		default:
			break;
	}
	x=item_x;
	y=item_y;
   /*These signals cause problems with my Click-And-Move detection*/
   if(event->type!=GDK_LEAVE_NOTIFY){
      gdk_event_free(last_event);
      last_event=gdk_event_copy(event);
   }
   return(retval);
}
static gint velocity_view_animation_cb(gpointer user_data){
	velocity_view*vview=(velocity_view*)user_data;
	if(vview->handle){
		if(vview->animation_handle){
			velocity_icon_set_animation_type(vview->animation_handle,vview->prefs->animation_mode);
			switch(vview->prefs->animation_mode){
				case VELOCITY_ICON_ANIMATION_BOUNCE:
					velocity_icon_animate(vview->animation_handle,4.0);
					break;
				case VELOCITY_ICON_ANIMATION_ROTATE:
					velocity_icon_animate(vview->animation_handle,10.0);
					break;
				case VELOCITY_ICON_ANIMATION_PULSATE:
					velocity_icon_animate(vview->animation_handle,0.05);
					break;
				case VELOCITY_ICON_ANIMATION_INVALID:
				default:
					break;
			}
		}
	}
	return(TRUE);
}
/*Tooltip info*/
static char*velocity_view_tooltip_name_cb(velocity_vfs_obj*vfsobj){
	return(g_strdup(vfsobj->info->name));
}
static char*velocity_view_tooltip_type_cb(velocity_vfs_obj*vfsobj){
	const char*md=gnome_vfs_mime_get_description(vfsobj->info->mime_type);
	if(md)return(g_strdup(md));
	return(g_strdup(_("No description")));
}
static char*velocity_view_tooltip_mime_cb(velocity_vfs_obj*vfsobj){
	return(g_strdup(vfsobj->info->mime_type));
}
static char*velocity_view_tooltip_size_cb(velocity_vfs_obj*vfsobj){
	return(gnome_vfs_format_file_size_for_display(velocity_view_last->info->size));
}
static void velocity_view_loading_cb(GnomeVFSAsyncHandle*handle,GnomeVFSResult result,GList*list,guint entries_read,gpointer user_data){
	velocity_view*vview=(velocity_view*)user_data;
	velocity_vfs_obj*vfsobj=NULL;
	GCompareFunc sortfunc=velocity_view_sort_table[vview->prefs->sorting];
	GList*n=list;		/*List iterator*/
	GList*dq;	/*Directory queue*/	/*These two may be the same*/
	GList*fq;	/*File queue*/
	if((result!=GNOME_VFS_OK)&&(result!=GNOME_VFS_ERROR_EOF)){
		//TODO:  ftp://   FTP get UN/PW here!
		//TODO:  smb://   ** WARNING **: Couldn't find a master browser
		char*msgstr=g_strdup_printf(_("Unable to read from \"%s\"!\nPlease try again and check that you have permission to access this object."),vview->clocation);
		GtkWidget*dlg=gtk_message_dialog_new(NULL,GTK_DIALOG_MODAL,GTK_MESSAGE_WARNING,GTK_BUTTONS_OK,msgstr);
		velocity_view_stop(vview);
		gtk_dialog_run(GTK_DIALOG(dlg));
		gtk_widget_destroy(dlg);
		g_free(msgstr);
		return;
	}
	if(vview->prefs->dirs_first){
		dq=vview->dirs;
		fq=vview->files;
	}else{
		dq=vview->files;
		fq=dq;
	}
	while(n){
		GnomeVFSFileInfo*vfsFileInfo=(GnomeVFSFileInfo*)n->data;
		gnome_vfs_file_info_ref(vfsFileInfo);
		if(!velocity_vfs_filter(vfsFileInfo,vview->shellexp,vview->prefs->filter)){
			/*Add file to the proper queue*/
			vfsobj=velocity_vfs_obj_new(vview->clocation,vfsFileInfo,NULL,vview);
			
			if(velocity_vfs_is_directory(vfsobj)){
				if(sortfunc){
					dq=g_list_insert_sorted(dq,(gpointer)vfsobj,sortfunc);
				}else{
					dq=g_list_prepend(dq,(gpointer)vfsobj);
				}
				if(!vview->prefs->dirs_first)fq=dq;	/*If we aren't putting dirs_first then we need to make sure these stay the same*/
				vview->total.dirs++;
			}else{
				if(sortfunc){
					fq=g_list_insert_sorted(fq,(gpointer)vfsobj,sortfunc);
				}else{
					fq=g_list_prepend(fq,(gpointer)vfsobj);
				}
				vview->total.files++;
				if(!vview->prefs->dirs_first)dq=fq;	/*If we aren't putting dirs_first then we need to make sure these stay the same*/
			}
			vview->total.objects++;
		}else{
			vview->read.filtered++;	/*Keep track of filtered items*/
		}
		gnome_vfs_file_info_unref(vfsFileInfo);
		n=n->next;
	}
	vview->files=fq;
	vview->dirs=dq;
	/*Check and see if we are done yet*/
	if(result==GNOME_VFS_ERROR_EOF){
		/*Since its not read we'll put the total both places to avoid confusion (and a sanity check too!)*/
		vview->read.filtered-=2;
		if(vview->read.filtered<0)vview->read.filtered=0;
		vview->total.filtered=vview->read.filtered;
		/*Start dequeing items*/
		velocity_view_dequeue_prep(vview);
		VELOCITY_MSG("done scanning");
		velocity_view_dequeue(vview);
	}else{
		gtk_main_iteration_do(FALSE);
	}
}
void velocity_view_set_icon_caption(velocity_view*vview,velocity_vfs_obj*vfsobj,gboolean confine){
	char*base=NULL;
	if(vview->on_behavior_event)base=(char*)vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_OBJECT_GET_CAPTION,NULL,vfsobj,vview->user_data);
	if(!base){
		const char*type=gnome_vfs_mime_get_description(vfsobj->info->mime_type);
		//TODO:  need to be able to get dir sizes
		char*size=NULL;
		if((vview->prefs->profile->caption_type==VELOCITY_VIEW_CAPTION_NAME_TYPE_SIZE)||
			(vview->prefs->profile->caption_type==VELOCITY_VIEW_CAPTION_NAME_SIZE_TYPE)||
			(vview->prefs->profile->caption_type==VELOCITY_VIEW_CAPTION_NAME_SIZE)){
			if(velocity_vfs_is_directory(vfsobj)){
				//TODO:  This needs to be async...
				GList*l=NULL;
				char*uri=velocity_vfs_get_uri(vfsobj);
				VELOCITY_MSG("get_dir_items(%s)",uri);
				if(gnome_vfs_directory_list_load(&l,uri,0)==GNOME_VFS_OK){
					size=g_strdup_printf(_("%i items"),MAX(g_list_length(l)-2,0));
					gnome_vfs_file_info_list_unref(l);
					g_list_free(l);
				}
				g_free(uri);
			}else{
				size=gnome_vfs_format_file_size_for_display(vfsobj->info->size);
			}
			if(!size)size=g_strdup(_("Unknown"));
		}
		switch(vview->prefs->profile->caption_type){
			case VELOCITY_VIEW_CAPTION_NAME_TYPE_SIZE:
				base=g_strconcat(vfsobj->info->name,"\n",type,"\n",size,NULL);
				break;
			case VELOCITY_VIEW_CAPTION_NAME_TYPE:
				base=g_strconcat(vfsobj->info->name,"\n",type,NULL);
				break;
			case VELOCITY_VIEW_CAPTION_NAME_SIZE_TYPE:
				base=g_strconcat(vfsobj->info->name,"\n",size,"\n",type,NULL);
				break;
			case VELOCITY_VIEW_CAPTION_NAME_SIZE:
				base=g_strconcat(vfsobj->info->name,"\n",size,NULL);
				break;
			case VELOCITY_VIEW_CAPTION_NAME:
			default:
				base=g_strdup(vfsobj->info->name);
				break;
		}
		g_free(size);
	}
	velocity_icon_set_caption(vfsobj->icon,base,vview->prefs->font,confine);
	g_free(base);
}
static void velocity_view_draw_background_cb(GnomeCanvas*canvas,GdkDrawable*drawable,int x,int y,int width,int height,gpointer user_data){
	velocity_view*vview=(velocity_view*)user_data;
	if(GTK_WIDGET_VISIBLE(vview->canvas)&&(x>=0)&&(y>=0)){
		GdkGCValuesMask value_mask;
		GdkGCValues gc_values;
		if(vview->background){
			gc_values.tile=vview->background;
			gc_values.ts_x_origin=-x;
			gc_values.ts_y_origin=-y;
			gc_values.fill=GDK_TILED;
			value_mask=GDK_GC_FILL|GDK_GC_TILE|GDK_GC_TS_X_ORIGIN|GDK_GC_TS_Y_ORIGIN;
		}else{
			gc_values.foreground=vview->prefs->background;
			gdk_rgb_find_color(gtk_widget_get_colormap(GTK_WIDGET(canvas)),&gc_values.foreground);
			gc_values.fill=GDK_SOLID;
			value_mask=GDK_GC_FILL|GDK_GC_FOREGROUND;
		}
		gdk_gc_set_values(canvas->pixmap_gc,&gc_values,value_mask);
		gdk_draw_rectangle(drawable,canvas->pixmap_gc,TRUE,0,0,width,height);

		/*Stop GnomeCanvas from interfering too much*/
		g_signal_stop_emission_by_name(canvas,"draw_background");
	}
}
//TODO:  This should be able to deal with stretch,center,tile,etc. rather than just stretch (as it is now)
void velocity_view_set_background(velocity_view*vview,const char*uri){
	GdkPixbuf*p;
	velocity_view_ensure_canvas(vview);
	if(vview->background)gdk_drawable_unref(GDK_DRAWABLE(vview->background));
	p=gdk_pixbuf_new_from_file((char*)uri,NULL);
	if(p){
		GdkPixbuf*p2=gdk_pixbuf_scale_simple(p,gdk_screen_width(),gdk_screen_height(),GDK_INTERP_TILES);
		if(p2){
			GtkStyle*style=gtk_widget_get_style(GTK_WIDGET(vview->canvas));
			gdk_pixbuf_render_pixmap_and_mask(p2,&vview->background,NULL,127);
			
			/*set new background style (keeps canvas things smoother looking)*/
			style->bg_pixmap[GTK_STATE_NORMAL]=vview->background;
			gtk_widget_set_style(GTK_WIDGET(vview->canvas),style);
			
			g_object_unref(p2);
		}
		g_object_unref(p);
	}
	g_signal_connect(vview->canvas,"draw_background",G_CALLBACK(velocity_view_draw_background_cb),vview);
}
void velocity_view_ensure_canvas(velocity_view*vview){
	if(!vview->canvas){
		if(vview->canvas){
			gtk_container_remove(GTK_CONTAINER(vview->widget),vview->canvas);
			gtk_object_destroy(GTK_OBJECT(vview->canvas)); /*destroy root canvas*/
		}
		/*Create a new root canvas*/
		vview->canvas=gnome_canvas_new();
		gnome_canvas_set_pixels_per_unit(GNOME_CANVAS(vview->canvas),1.0);
		gnome_canvas_set_center_scroll_region(GNOME_CANVAS(vview->canvas),FALSE);
		gtk_container_add(GTK_CONTAINER(vview->widget),vview->canvas);
		//TODO:  User settable background	AND a solid color background (easiest and fastest)
		velocity_view_set_background_color(vview,&vview->prefs->background);
		GNOME_CANVAS(vview->canvas)->close_enough=8;
		gtk_widget_show(vview->canvas);
		g_signal_connect(vview->canvas,"event",(GtkSignalFunc)velocity_view_canvas_cb,vview);
		/*Configure our DND*/
		if(vview->prefs->can_drag){
			gtk_drag_dest_set(GTK_WIDGET(vview->canvas),GTK_DEST_DEFAULT_ALL,velocity_dnd_targets,VELOCITY_DND_TARGETS_NUM,GDK_ACTION_DEFAULT|GDK_ACTION_COPY|GDK_ACTION_LINK);
		}
		gnome_canvas_scroll_to(GNOME_CANVAS(vview->canvas),0,0);
		gtk_scrolled_window_set_vadjustment(GTK_SCROLLED_WINDOW(vview->widget),NULL);
		gtk_scrolled_window_set_hadjustment(GTK_SCROLLED_WINDOW(vview->widget),NULL);
		g_signal_connect(vview->canvas,"drag_begin",G_CALLBACK(velocity_view_drag_begin_cb),vview);
		g_signal_connect(vview->canvas,"drag_data_get",G_CALLBACK(velocity_view_drag_get_cb),vview);
		g_signal_connect(vview->canvas,"drag_motion",G_CALLBACK(velocity_view_drag_motion_cb),vview);
		g_signal_connect(vview->canvas,"drag_data_received",G_CALLBACK(velocity_view_drag_received_cb),vview);
		g_signal_connect(vview->canvas,"size_allocate",(GtkSignalFunc)velocity_view_resize_cb,vview);
		velocity_view_resize_idle_cb((gpointer)vview);	/*Forcefully refresh...this is only needed in very rare cases such as this...*/
	}
	if(!vview->canvas_group){
		/*make a new canvas group as to clear the canvas easily (by nuke the canvas group)*/
		vview->canvas_group=GNOME_CANVAS_GROUP(gnome_canvas_item_new(gnome_canvas_root(GNOME_CANVAS(vview->canvas)),gnome_canvas_group_get_type(),
			"x",(gdouble)0.0, "y",(gdouble)0.0, NULL));
		gnome_canvas_set_scroll_region(GNOME_CANVAS(vview->canvas),0.0,0.0,0.0,0.0);
	}
}
static void velocity_view_icon_inc_cords(velocity_view*vview){
	/*Drawing direction method*/
	//TODO:  Finish and optimize this code
	if((vview->prefs->profile->draw_start==GTK_ANCHOR_NW)&&(vview->prefs->profile->draw_direction==GTK_ANCHOR_EAST)){
		vview->icon.x+=vview->prefs->profile->icon_spacing.x;
		if(vview->icon.x>=(vview->widget->allocation.width-vview->prefs->profile->icon_spacing.x-vview->border.left-vview->border.right)){
			vview->icon.x=(vview->prefs->profile->icon_spacing.x-vview->prefs->profile->icon_size.x)>>1+vview->border.left;
			vview->icon.y+=vview->prefs->profile->icon_spacing.y;
		}
	}else if((vview->prefs->profile->draw_start==GTK_ANCHOR_NW)&&(vview->prefs->profile->draw_direction==GTK_ANCHOR_SOUTH)){
		vview->icon.y+=vview->prefs->profile->icon_spacing.y;
		if(vview->icon.y>=((vview->widget->allocation.height)-vview->prefs->profile->icon_spacing.y-vview->border.top-vview->border.bottom)){
			vview->icon.y=vview->border.top;
			vview->icon.x+=vview->prefs->profile->icon_spacing.x;
		}
	}else if((vview->prefs->profile->draw_start==GTK_ANCHOR_NE)&&(vview->prefs->profile->draw_direction==GTK_ANCHOR_SOUTH)){
		vview->icon.y+=vview->prefs->profile->icon_spacing.y;
		if(vview->icon.y>=((vview->widget->allocation.height)-vview->prefs->profile->icon_spacing.y-vview->border.top-vview->border.bottom)){
			vview->icon.y=vview->border.top;
			vview->icon.x-=vview->prefs->profile->icon_spacing.x;
		}
	}
}
static gint velocity_view_dequeue_cb(velocity_view*vview){
	velocity_vfs_obj*vfsobj=NULL;
	GList*n=NULL;
	guint i;
	guint items=vview->total.objects-vview->read.objects;	/*Avoid too much drawing w/o returning to the main gtk loop*/
	if((vview->dequeue_handle)&&(items>VELOCITY_VIEW_RENDER_PER_SYNC_LOOP))items=VELOCITY_VIEW_RENDER_PER_SYNC_LOOP;
	if(vview->handle){
		gnome_vfs_async_cancel(vview->handle);
		vview->handle=NULL;
		if(vview->animation_idle_handle){
			gtk_timeout_remove(GPOINTER_TO_INT(vview->animation_idle_handle));
			vview->animation_idle_handle=(guint)NULL;
		}
		vview->animation_handle=(guint)NULL;
		/*nuke our old icons here*/
		n=vview->nuke_files;
		while(n){
			vfsobj=(velocity_vfs_obj*)n->data;
			if(vfsobj){
				/*Force the icon to be NULL so nobody tries to use it*/
				vfsobj->icon=NULL;
				velocity_vfs_obj_free(vfsobj);
			}
			n=n->next;
		}
		if(vview->nuke_dirs!=vview->nuke_files){
			n=vview->nuke_dirs;
			while(n){
				vfsobj=(velocity_vfs_obj*)n->data;
				if(vfsobj){
					/*Force the icon to be NULL in case somebody else was hoping to use it*/
					//if(vfsobj->icon)velocity_icon_free(vfsobj->icon);
					vfsobj->icon=NULL;
					velocity_vfs_obj_free(vfsobj);
				}
				n=n->next;
			}
			if(vview->nuke_dirs){
				g_list_free(vview->nuke_dirs);
				vview->nuke_dirs=NULL;
			}
		}
		if(vview->nuke_files){
			g_list_free(vview->nuke_files);
			vview->nuke_files=NULL;
		}
		VELOCITY_MSG("Clearing view");
		/*clear our canvas*/
		if((vview->canvas)&&(vview->canvas_group)){
			gtk_object_destroy(GTK_OBJECT(vview->canvas_group));
			velocity_view_canvas_cb(NULL,NULL,(gpointer)vview);
			vview->canvas_group=NULL;
		}
		velocity_view_ensure_canvas(vview);
		gnome_canvas_scroll_to(GNOME_CANVAS(vview->canvas),0,0);
		gtk_scrolled_window_set_vadjustment(GTK_SCROLLED_WINDOW(vview->widget),NULL);
		gtk_scrolled_window_set_hadjustment(GTK_SCROLLED_WINDOW(vview->widget),NULL);
	}
	if(items==0)goto velocity_view_dequeue_fini;
	for(i=0;i<items;i++){
		if(vview->read.objects<vview->total.objects){
			if(!vview->qn){	/*This is the first time around or we are done with the dirs*/
				if(vview->prefs->dirs_first){
					if(vview->read.dirs==vview->total.dirs){
						vview->qn=vview->files;
					}else{
						vview->qn=vview->dirs;
					}
				}else{
					vview->qn=vview->files;
				}
			}
			if(vview->qn){
				char*emblem=NULL;
				vfsobj=(velocity_vfs_obj*)vview->qn->data;

				if(vfsobj->icon){
					/*If we are resorting then we can just move icons around*/
					velocity_icon_move_to(vfsobj->icon,vview->icon.x,vview->icon.y);
				}else{
					/*Deal with items that don't have MIME info*/
					if((!vfsobj->info->mime_type)||(strcmp(vfsobj->info->mime_type,GNOME_VFS_MIME_TYPE_UNKNOWN)==0))velocity_view_force_info_update(vview,vfsobj);

					/*Install basic tooltip*/
					velocity_vfs_obj_ext_add(vfsobj,VELOCITY_TOOLTIPS_EXT_ID,NULL,NULL);
					velocity_tooltip_add(vfsobj,"name",_("Name"),VELOCITY_TOOLTIP_ITEM_TYPE_TEXT,(gpointer)velocity_view_tooltip_name_cb);
					velocity_tooltip_add(vfsobj,"type",_("Type"),VELOCITY_TOOLTIP_ITEM_TYPE_TEXT,(gpointer)velocity_view_tooltip_type_cb);
					velocity_tooltip_add(vfsobj,"mime",_("MIME"),VELOCITY_TOOLTIP_ITEM_TYPE_TEXT,(gpointer)velocity_view_tooltip_mime_cb);
					velocity_tooltip_add(vfsobj,"size",_("Size"),VELOCITY_TOOLTIP_ITEM_TYPE_TEXT,(gpointer)velocity_view_tooltip_size_cb);
					if(vview->on_behavior_event)vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_OBJECT_LOADED,NULL,vfsobj,vview->user_data);

					/*Setup a template basic_icon w/o an icon or a caption*/
					vfsobj->icon=velocity_icon_new(GNOME_CANVAS_GROUP(vview->canvas_group),vview->icon.x,vview->icon.y,
						vview->prefs->profile->icon_size,vview->prefs->profile->icon_spacing,vview->prefs->profile->caption_placement);
					velocity_icon_set_font_colors(vfsobj->icon,
						gdk_color_to_rgba(vview->prefs->caption_fg,255),gdk_color_to_rgba(vview->prefs->caption_bg,192),
						gdk_color_to_rgba(vview->prefs->caption_pfg,255),gdk_color_to_rgba(vview->prefs->caption_pbg,192),
						gdk_color_to_rgba(vview->prefs->caption_hfg,255),gdk_color_to_rgba(vview->prefs->caption_hbg,192));
					velocity_icon_set_selector_type(vfsobj->icon,vprefs->selector_type);
					velocity_view_set_icon_caption(vview,vfsobj,TRUE);	/*confine caption text by default*/
					if(vview->on_behavior_event){	/*execute gui hooks*/
						GdkPixbuf*r=vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_OBJECT_GET_ICON,NULL,vfsobj,vview->user_data);
						if(r){
							velocity_icon_set_icon(vfsobj->icon,r);
							g_object_unref(r);
						}
					}
					/*Add emblems here*/
					/*Broken Symlink detection*/
					if(velocity_vfs_is_symlink(vfsobj)){
						GnomeVFSURI*guri=NULL;
						emblem=VELOCITY_EMBLEM_SYMLINK;
						if(vfsobj->info->symlink_name[0]==GNOME_VFS_URI_PATH_CHR){
							guri=gnome_vfs_uri_new(vfsobj->info->symlink_name);
						}else{
							guri=gnome_vfs_uri_new(vview->clocation);
							guri=gnome_vfs_uri_append_string(guri,vfsobj->info->symlink_name);
						}
						if(!gnome_vfs_uri_exists(guri)){
							VELOCITY_MSG("broken symlink: %s->%s",guri->text,vfsobj->info->symlink_name);
							emblem=VELOCITY_EMBLEM_BROKEN_SYMLINK;
						}
						gnome_vfs_uri_unref(guri);
					}
					velocity_icon_emblem_add(vfsobj->icon,emblem);
					if(!velocity_vfs_is_readable(vfsobj))velocity_icon_emblem_add(vfsobj->icon,VELOCITY_EMBLEM_NOREAD);
					if(!vfsobj->icon->pixbuf){
						GdkPixbuf*p=velocity_view_get_mime_icon(vview,vfsobj,vview->prefs->profile->icon_size.x,vview->prefs->profile->icon_size.y);
						if(p){
							velocity_icon_set_icon(vfsobj->icon,p);
							g_object_unref(p);
						}
					}
					if(!vfsobj->icon->pixbuf){	/*this should not happen*/
						VELOCITY_ERR(_("Unable to find icons.  Please reinstall GNOME themes and/or Velocity."));
						return(TRUE);
					}
					g_signal_connect(vfsobj->icon->icon_group,"event",(GtkSignalFunc)velocity_view_icon_cb,vfsobj);
				}
				if(velocity_vfs_is_directory(vfsobj)){
					vview->read.dirs++;
				}else{
					vview->read.files++;
				}
				velocity_view_icon_inc_cords(vview);
				vview->read.objects++;
				vview->qn=vview->qn->next;
			}
		}else{
velocity_view_dequeue_fini:
			if(vview->dequeue_handle){
				vview->dequeue_handle=(guint)NULL;
				if(vview->read.objects+vview->read.images==vview->total.objects+vview->total.images){	/*hey, it could happen...*/
					if(vview->on_status_event)vview->on_status_event(vview,VELOCITY_VIEW_STATUS_EVENT_DONE,vview->clocation,GNOME_VFS_OK,vview->user_data);
				}
				return(FALSE);
			}
		}
	}
	velocity_view_resize_cb(NULL,NULL,(gpointer)vview);
	return(TRUE);
}
void velocity_view_loading_prep(velocity_view*vview){
	vview->total.files=vview->total.dirs=vview->total.objects=vview->total.images=vview->total.filtered=0;
	/*Move lists around so that they can be nuked during dequeing and the canvas redraw*/
	vview->nuke_dirs=vview->dirs;
	vview->nuke_files=vview->files;
	vview->dirs=vview->files=(guint)NULL;
	velocity_view_last=NULL;
}
void velocity_view_dequeue_prep(velocity_view*vview){
	vview->read.files=vview->read.dirs=vview->read.objects=vview->read.images=vview->read.filtered=0;
	vview->qn=NULL;
	if(vview->prefs->profile->draw_start==GTK_ANCHOR_NW){
		switch(vview->prefs->profile->caption_placement){
			case GTK_ANCHOR_SOUTH:
				vview->icon.x=((vview->prefs->profile->icon_spacing.x-vview->prefs->profile->icon_size.x)>>1)+vview->border.left;
				vview->icon.y=vview->border.top;
				break;
			case GTK_ANCHOR_EAST:
			default:
				vview->icon.x=vview->border.left;
				vview->icon.y=vview->border.top;
				break;
		}
	}else if(vview->prefs->profile->draw_start==GTK_ANCHOR_NE){
		switch(vview->prefs->profile->caption_placement){
			case GTK_ANCHOR_SOUTH:
				vview->icon.x=vview->widget->allocation.width-((vview->prefs->profile->icon_spacing.x+vview->prefs->profile->icon_size.x)>>1)-vview->border.left;
				vview->icon.y=vview->border.top;
				break;
		}			
	}
	if(vview->tooltip_handle)gtk_timeout_remove(vview->tooltip_handle);
	if(velocity_view_last)velocity_vfs_obj_free(velocity_view_last);
	if(vview->refresh_handle)gtk_idle_remove(vview->refresh_handle);
	velocity_view_last=NULL;
	vview->tooltip_handle=vview->refresh_handle=(guint)NULL;
	velocity_clipboard_flush(vview->selection);
}
static void velocity_view_validated_cb(GnomeVFSAsyncHandle*handle,GList*results,gpointer user_data){
	velocity_view*vview=(velocity_view*)user_data;
	velocity_vfs_obj*vfsobj=NULL;
	char*uri=NULL;
	vview->lock=FALSE;	/*Should be safe now*/
	if((results)&&(results->data)){
		GnomeVFSGetFileInfoResult*r=(GnomeVFSGetFileInfoResult*)results->data;
		if(vview->shellexp){
			g_free(vview->shellexp);
			vview->shellexp=NULL;
		}
		uri=gnome_vfs_uri_to_string(r->uri,GNOME_VFS_URI_HIDE_NONE);
		if(r->result==GNOME_VFS_OK){ /*It exists*/
			vfsobj=velocity_vfs_obj_new(uri,r->file_info,NULL,(gpointer)vview);
			if(vfsobj){
				/*I had this issue so I am taking care of it right now...*/
				//TODO:  This causes problems and is inconsistant...figure out where the real problem is and nuke this _please_!
				if(!velocity_vfs_is_readable(vfsobj))r->result=GNOME_VFS_ERROR_ACCESS_DENIED;
				if((velocity_vfs_is_directory(vfsobj))&&(r->result==GNOME_VFS_OK)){
					char*tmp=uri;
					uri=velocity_vfs_ensure_dir_uri(tmp);
					g_free(tmp);
				}else{
					if(vview->on_status_event)vview->on_status_event(vview,VELOCITY_VIEW_STATUS_EVENT_ERROR,uri,r->result,vview->user_data);
					velocity_vfs_obj_free(vfsobj);
					return;
				}
				velocity_vfs_obj_free(vfsobj);
			}
		}else{
			char*tail;
			if(r->result==GNOME_VFS_ERROR_LOGIN_FAILED){
				VELOCITY_MSG("Rewrite Password dialog and put here...");
			}
			//TODO: if not found and starts with www. add http://, ftp. add ftp://
			/*Shell expression handling*/ 
			tail=velocity_vfs_uri_get_tail(uri);
			if((strchr(tail,'*'))||(strchr(tail,'?'))||(strchr(tail,'['))||(strchr(tail,']'))){	//TODO:  others?  unfinished?
				char*lastdelim=strrchr(uri,'/');
				char*newuri=g_strndup(uri,lastdelim+1-uri);
				vview->shellexp=g_strdup(tail);
				VELOCITY_MSG("shellexp=%s, uri=%s",vview->shellexp,newuri);
				velocity_view_goto(vview,newuri);
				g_free(newuri);
			}else{
				if(vview->on_status_event)vview->on_status_event(vview,VELOCITY_VIEW_STATUS_EVENT_ERROR,uri,r->result,vview->user_data);
			}
			g_free(tail);
			return;
		}
	}
	gnome_vfs_async_cancel(vview->handle);
	vview->handle=NULL;
	if(uri){
		velocity_view_goto(vview,uri);
		g_free(uri);
	}
}
static void velocity_view_goto(velocity_view*vview,const char*uri){
	/*We might as well go ahead and set this now because we will be scanning in a couple of clocks...*/
	velocity_view_loading_prep(vview);

	if(vview->clocation)g_free(vview->clocation);
	vview->clocation=velocity_vfs_uri_reformat(uri);	/*keep local if possible*/

	if(vview->on_status_event)vview->on_status_event(vview,VELOCITY_VIEW_STATUS_EVENT_GO,uri,GNOME_VFS_OK,vview->user_data);

	VELOCITY_MSG("goto(%s)",uri);
	gnome_vfs_async_load_directory(&vview->handle,uri,
		GNOME_VFS_FILE_INFO_DEFAULT|GNOME_VFS_FILE_INFO_GET_MIME_TYPE|GNOME_VFS_FILE_INFO_FORCE_FAST_MIME_TYPE|GNOME_VFS_FILE_INFO_FOLLOW_LINKS,
		VELOCITY_VIEW_LOAD_PER_SYNC_LOOP*4,GNOME_VFS_PRIORITY_DEFAULT,
		velocity_view_loading_cb,(gpointer)vview);
}
static void velocity_view_validate(velocity_view*vview,const char*uri){
	GnomeVFSURI*guri=gnome_vfs_uri_new(uri);
	if(guri){
		char*turi=gnome_vfs_uri_to_string(guri,GNOME_VFS_URI_HIDE_NONE);
		if(turi){
			GList*uris=g_list_prepend(NULL,guri);
			if(vview->on_status_event)vview->on_status_event(vview,VELOCITY_VIEW_STATUS_EVENT_VALIDATE,turi,GNOME_VFS_OK,vview->user_data);
			gnome_vfs_async_get_file_info(&vview->handle,uris,GNOME_VFS_FILE_INFO_DEFAULT,GNOME_VFS_PRIORITY_DEFAULT,
				(GnomeVFSAsyncGetFileInfoCallback)velocity_view_validated_cb,(gpointer)vview);
			vview->lock=TRUE;	/*We want to set our lock while validating*/
			g_list_free(uris);
			g_free(turi);
		}
		gnome_vfs_uri_unref(guri);
	}else{
		if(vview->on_status_event)vview->on_status_event(vview,VELOCITY_VIEW_STATUS_EVENT_ERROR,uri,GNOME_VFS_ERROR_INVALID_URI,vview->user_data);
	}
}
void velocity_view_set_location(velocity_view*vview,const char*uri){
	char*uri2=(uri==NULL)?g_strdup(vprefs->locations->home):g_strdup(uri);
	velocity_view_stop(vview);	/*stop anything else going on*/
	vview->animation_idle_handle=gtk_timeout_add(25,velocity_view_animation_cb,(gpointer)vview);	/*Start the animating icon*/
	velocity_view_validate(vview,uri2);
	g_free(uri2);
}
void velocity_view_soft_refresh(velocity_view*vview,gboolean reload_images){
	/*NOTE: This won't work for the "Don't Sort" setting because the order is already messed up from previous sortings*/
	velocity_vfs_obj*vfsobj=NULL;
	GCompareFunc sortfunc=velocity_view_sort_table[vview->prefs->sorting];
	GList*dq=vview->dirs;
	GList*fq=vview->files;
	GList*n=dq;
	if(vview->lock)return;
	velocity_view_stop(vview);
	vview->dirs=vview->files=NULL;
	while(n){
		if(n->data){
			vfsobj=(velocity_vfs_obj*)n->data;
			if(vfsobj->icon){
				if((reload_images)&&(velocity_thumbnail_can_thumbnail(vfsobj))){
					velocity_icon_free(vfsobj->icon);
					vfsobj->icon=NULL;
				}
			}
			if(sortfunc){
				vview->dirs=g_list_insert_sorted(vview->dirs,(gpointer)vfsobj,sortfunc);
			}else{
				vview->dirs=g_list_prepend(vview->dirs,(gpointer)vfsobj);
			}
		}
		n=n->next;
	}
	if(dq!=fq){
		n=fq;
		while(n){
			if(n->data){
				vfsobj=(velocity_vfs_obj*)n->data;
				if(vfsobj->icon){
					if((reload_images)&&(velocity_thumbnail_can_thumbnail(vfsobj))){
						velocity_icon_free(vfsobj->icon);
						vfsobj->icon=NULL;
					}
				}
				if(sortfunc){
					vview->files=g_list_insert_sorted(vview->files,(gpointer)vfsobj,sortfunc);
				}else{
					vview->files=g_list_prepend(vview->files,(gpointer)vfsobj);
				}
				n=n->next;
			}
		}
		g_list_free(fq);
	}else{
		vview->files=vview->dirs;
	}
	g_list_free(dq);
	velocity_view_dequeue_prep(vview);
	vview->dequeue_handle=gtk_idle_add_priority(GTK_PRIORITY_RESIZE+1,(GtkFunction)velocity_view_dequeue_cb,(gpointer)vview);
	gtk_widget_queue_draw(GTK_WIDGET(vview->canvas));	/*In case the GtkStyle has been updated or something*/
}
//TODO:  This should be activate_with(view,progname) and if !progname follow defaults
/*void velocity_view_activate_with_app(velocity_view*vview,GnomeVFSMimeApplication*app){
	
}
void velocity_view_activate_with_custom(velocity_view*vview,char*custom){
}*/
void velocity_view_activate(velocity_view*vview){
	GList*n=vview->selection->items;
	while(n){
		if(n->data){
			velocity_vfs_obj*vfsobj=(velocity_vfs_obj*)n->data;
			gpointer ret=NULL;
			if(vview->on_behavior_event)ret=vview->on_behavior_event(vview,VELOCITY_VIEW_BEHAVIOR_EVENT_ACTIVATE,NULL,vfsobj,vview->user_data);
			if((!ret)&&(vview->prefs->can_open))velocity_view_open(vview,vfsobj);
		}
		n=n->next;
	}
}
//TODO:  This should be open_with(view,progname) and if !progname follow defaults
void velocity_view_open(velocity_view*vview,velocity_vfs_obj*vfsobj){
	GnomeVFSMimeApplication*vfsmimeapp=NULL;
	if((!vfsobj)||(!vfsobj->info)||(!vfsobj->info->name))return;
	if(velocity_vfs_is_directory(vfsobj)){
		char*uri=velocity_vfs_get_uri_unescaped(vfsobj);
		if((!vview->prefs->dirs_in_same_win)&&(vview->on_status_event)){	/*open in new window*/
			VELOCITY_MSG("calling new window callback");
			/*HACK:  This is a hack to avoid an invalid lasso drag when opening a new view/window*/
			velocity_view_clicked=FALSE;
			velocity_view_click.x=velocity_view_click.y=0;
			vview->on_status_event(vview,VELOCITY_VIEW_STATUS_EVENT_NEW_WINDOW,uri,GNOME_VFS_OK,vview->user_data);
		}else{
			vview->animation_handle=vfsobj->icon;
			if((!velocity_vfs_is_trash_dir(vfsobj))&&(!velocity_vfs_is_home_dir(vfsobj))){
				char*iuri=velocity_image_get_uri(VELOCITY_FOLDER_ACCEPT,MAX(vview->prefs->profile->icon_size.x,vview->prefs->profile->icon_size.y));
				GdkPixbuf*pb=velocity_image_cache_get(iuri,vview->prefs->profile->icon_size.x,vview->prefs->profile->icon_size.y,NULL,VELOCITY_IMAGE_CACHE_PRIORITY_HIGH);
				if(pb){
					velocity_icon_set_icon(vfsobj->icon,pb);
					g_object_unref(pb);
				}
				g_free(iuri);
			}
			velocity_view_set_location(vview,uri);
		}
		g_free(uri);
	}else{
		char*uri=velocity_vfs_get_uri_unescaped(vfsobj);
		if((strncasecmp(vfsobj->info->mime_type,"text/",5)==0)&&(velocity_vfs_is_executable(vfsobj))){
			char*msgstr=g_strdup_printf(_("'%s' is a text file, but has executable permissions.\nWould you rather execute or open this file?"),vfsobj->info->name);
			GtkWidget*dlg=gtk_message_dialog_new(NULL,GTK_DIALOG_MODAL,GTK_MESSAGE_QUESTION,GTK_BUTTONS_NONE,msgstr);
			int a;
			gtk_dialog_add_buttons(GTK_DIALOG(dlg),_("Execute"),0,_("Open as Text"),1,NULL);
			a=gtk_dialog_run(GTK_DIALOG(dlg));
			gtk_widget_destroy(dlg);
			g_free(msgstr);
			if(a==0){
				char*argstr=g_strdup_printf("\"%s\" -e \"%s\"",vprefs->locations->terminal,uri);
				g_spawn_command_line_async(argstr,NULL);
				g_free(argstr);
				g_free(uri);
				return;
			}
		}
		vfsmimeapp=gnome_vfs_mime_get_default_application(vfsobj->info->mime_type);
		if(strcasecmp(vfsobj->info->mime_type,"application/x-executable-binary")==0){
			char*argstr=g_strdup_printf("\"%s\"",uri);
			g_spawn_command_line_async(argstr,NULL);
			g_free(argstr);
		}else{
			if(vfsmimeapp){
				if(vfsmimeapp->command){
					char*argstr=g_strdup_printf("%s \"%s\"",vfsmimeapp->command,uri);
					g_spawn_command_line_async(argstr,NULL);
					g_free(argstr);
				}
			}
		}
		g_free(uri);
	}
}
void velocity_view_stop(velocity_view*vview){
	if(vview->handle){
		gnome_vfs_async_cancel(vview->handle);
		vview->handle=NULL;
	}
	if(vview->dequeue_handle){
		gtk_idle_remove(vview->dequeue_handle);
		vview->dequeue_handle=(guint)NULL;
	}
	if(vview->animation_idle_handle){
		if(vview->animation_handle)velocity_icon_reset_animation(vview->animation_handle);
		gtk_timeout_remove(vview->animation_idle_handle);
		vview->animation_idle_handle=(guint)NULL;
	}
	velocity_view_tooltip_off(vview);
	velocity_view_hide_lasso(vview);
	velocity_thumbnail_free_all_by_data((gpointer)vview);	/*This kills all active loading threads and frees resources*/
	if(vview->on_status_event)vview->on_status_event(vview,VELOCITY_VIEW_STATUS_EVENT_STOP,vview->clocation,GNOME_VFS_OK,vview->user_data);
}
void velocity_view_show_lasso(velocity_view*vview,gfloat x1,gfloat y1,gfloat x2,gfloat y2){
	//TODO:  Colors (and stipple?) should be from prefs..
	if(!vview->lasso){
		char mystipple[]={170,(char)NULL};
		GdkBitmap*stipple=gdk_bitmap_create_from_data(NULL,mystipple,4,2);
		vview->lasso=velocity_lasso_new(GNOME_CANVAS_GROUP(vview->canvas_group),x1,y1,x2,y2,0x4444FF22,stipple,0x00006622);
	}else{
		velocity_lasso_moveto(vview->lasso,x1,y1,x2,y2);
	}
}
void velocity_view_hide_lasso(velocity_view*vview){
	if(vview->lasso){
		velocity_lasso_free(vview->lasso);
		vview->lasso=NULL;
	}
}
void velocity_view_clear_selection(velocity_view*vview){
	velocity_vfs_obj*vfsobj=NULL;
	GList*n=vview->selection->items;
	while(n){
		vfsobj=(velocity_vfs_obj*)n->data;
		if(vfsobj){
			if(vfsobj->icon){
				velocity_view_set_icon_caption(vview,vfsobj,TRUE);
				if(vfsobj->icon->highlighted)velocity_icon_unhighlight(vfsobj->icon);
				if(vfsobj->icon->prelighted)velocity_icon_unprelight(vfsobj->icon);
			}
		}
		n=n->next;
	}
	if(velocity_view_last)velocity_vfs_obj_free(velocity_view_last);
	velocity_view_last=NULL;
	velocity_clipboard_flush(vview->selection);
}
void velocity_view_foreach(velocity_view*vview,GFunc func,gpointer user_data){
	GList*n=vview->dirs;
	while(n){
		if(n->data)func(n->data,user_data);
		n=n->next;
	}
	if(vview->dirs!=vview->files){
		n=vview->files;
		while(n){
			if(n->data)func(n->data,user_data);
			n=n->next;
		}
	}
}
static void velocity_view_free_cb(GtkObject*obj,gpointer user_data){
	if(user_data){
		velocity_view*vview=(velocity_view*)user_data;
		velocity_view_free(vview);
	}
}
static void velocity_view_free_foreach(velocity_vfs_obj*vfsobj,gpointer user_data){
	if(vfsobj->icon){
		velocity_icon_free(vfsobj->icon);
		vfsobj->icon=NULL;
	}
	velocity_vfs_obj_free(vfsobj);
}
void velocity_view_free(velocity_view*vview){
	vview->on_status_event=NULL;
	vview->on_behavior_event=NULL;
	if(vview){
		GList*n=vview->dirs;
		velocity_view_stop(vview);
		velocity_view_foreach(vview,(GFunc)velocity_view_free_foreach,NULL);
		if(vview->files!=vview->dirs)g_list_free(vview->files);
		g_list_free(vview->dirs);
		if(vview->selection)velocity_clipboard_free(vview->selection);
		if(vview->widget)gtk_widget_destroy(vview->widget);
		g_free(vview);
		VELOCITY_MSG("object freed");
	}
}

GList*velocity_tooltip_items(velocity_vfs_obj*vfsobj){
	if(vfsobj){
		return((GList*)velocity_vfs_obj_ext_get_data(vfsobj,VELOCITY_TOOLTIPS_EXT_ID));
	}
	return(NULL);
}
void velocity_tooltip_add(velocity_vfs_obj*vfsobj,const char*id,const char*desc,velocity_tooltip_item_type type,gpointer data){
	GList*tips=velocity_tooltip_items(vfsobj);
	velocity_tooltip_item*i=g_new0(velocity_tooltip_item,1);
	i->id=g_strdup(id);
	i->desc=g_strdup(desc);
	i->type=type;
	i->data=data;
	tips=g_list_append(tips,(gpointer)i);
	velocity_vfs_obj_ext_set_data(vfsobj,VELOCITY_TOOLTIPS_EXT_ID,(gpointer)tips);
}
void velocity_tooltip_remove(velocity_vfs_obj*vfsobj,const char*id){
	GList*tips=velocity_tooltip_items(vfsobj);
	GList*n=tips;
	GList*nl=NULL;
	while(n){
		if(n->data){
			velocity_tooltip_item*i=(velocity_tooltip_item*)n->data;
			if(strcmp(i->id,id)==0){
				g_free(i->id);
				g_free(i->desc);
				g_free(i);
			}else{
				nl=g_list_append(nl,n->data);
			}
		}
		n=n->next;
	}
	g_list_free(tips);
	velocity_vfs_obj_ext_set_data(vfsobj,VELOCITY_TOOLTIPS_EXT_ID,(gpointer)nl);
}
